"""Command line interface for python script to compare MDB models using fuzzy & nlp matching"""

from pathlib import Path
from typing import List, Optional, Tuple

import click
import pandas as pd
import spacy
from bento_meta.mdb import SearchableMDB
from bento_meta.mdb.mdb_tools import ToolsMDB
from bento_meta.objects import Property


def _get_prop_fuzzy_matches(smdb, qstring) -> Optional[List[dict]]:
    """
    returns list of fuzzy matching property nodes

    :param smdb
    :param qstring
    """
    items = smdb.search_entity_handles(qstring)
    fuzzy_matches = []
    if not items:
        return fuzzy_matches
    for itm in items["properties"]:
        fuzzy_matches.append(
            {
                "nanoid": itm["ent"]["nanoid"],
                "model": itm["ent"]["model"],
                "handle": itm["ent"]["handle"],
            }
        )
    return fuzzy_matches


def _get_fuzzy_synonyms(mdb: ToolsMDB, fuzzy_match_list) -> list:
    """convert results of get_property_synonyms_all() to list of dicts"""
    fuzzy_synonym_list = []
    for fuzzy_match in fuzzy_match_list:
        fuzzy_synonyms = mdb.get_property_synonyms_all(
            Property({"nanoid": fuzzy_match["nanoid"]})
        )
        if not fuzzy_synonyms:
            continue
        for synonym in fuzzy_synonyms:
            fuzzy_synonym_list.append(
                {
                    "nanoid": synonym.nanoid,
                    "model": synonym.model,
                    "handle": synonym.handle,
                }
            )
    return fuzzy_synonym_list


def _join_model_fuzzy_matches(base_prop_dict, comp_model) -> str:
    """takes a comp model & row of base_node_prop and returns string with joined fuzzy matches"""
    match_list = []
    for fuzzy_synonym in base_prop_dict["fuzzy_matches"]:
        if fuzzy_synonym["model"] == comp_model:
            for node_handle in fuzzy_synonym["node_handles"]:
                match_list.append(f"{node_handle}.{fuzzy_synonym['handle']}")
    if not match_list:
        return ""
    if len(match_list) == 1:
        return match_list[0]
    return "|".join(match_list)


def _join_model_nlp_matches(base_prop_dict, comp_model) -> str:
    """takes a comp model & row of base_node_prop and returns string with joined nlp matches"""
    match_list = []
    for nlp_synonym in base_prop_dict["nlp_matches"]:
        if nlp_synonym["model"] == comp_model:
            round_sim = round(nlp_synonym["similarity"], 2)
            match_list.append(
                f"{nlp_synonym['node_handle']}.{nlp_synonym['handle']} ({round_sim})"
            )
    if not match_list:
        return ""
    elif len(match_list) == 1:
        return match_list[0]
    return "|".join(match_list)


@click.command()
@click.option(
    "--mdb_uri", required=True, type=str, prompt=True, help="metamodel database URI"
)
@click.option(
    "--mdb_user",
    required=True,
    type=str,
    prompt=True,
    help="metamodel database username",
)
@click.option(
    "--mdb_pass",
    required=True,
    type=str,
    prompt=True,
    help="metamodel database password",
)
@click.option(
    "--base_model",
    required=True,
    type=str,
    prompt=True,
    help="model to compare to other models in MDB",
)
@click.option(
    "--comp_model",
    required=True,
    type=str,
    prompt=True,
    multiple=True,
    default=["DST", "PDC", "GDC", "HTAN", "ICDC", "CCDI", "CDA", "CDS"],
    help="model names to compare to base model nodes/props",
)
@click.option(
    "--output_filepath",
    required=True,
    type=str,
    prompt=True,
    help="path to .csv file for output",
)
@click.option(
    "--sim_threshold",
    required=True,
    default=0.8,
    type=float,
    prompt=False,
    help="similarity threshold synonyms must meet to be considered. default = 0.8",
)
@click.option(
    "--num_nlp",
    required=True,
    default=3,
    type=int,
    prompt=False,
    help="similarity threshold synonyms must meet to be considered. default = 0.8",
)
def main(
    mdb_uri: str,
    mdb_user: str,
    mdb_pass: str,
    base_model: str,
    comp_model: Tuple[str],
    output_filepath: str,
    sim_threshold: float = 0.8,
    num_nlp: int = 3,
) -> None:
    """
    Compare base_model's node/property handles to those of comp_model(s)
    using fuzzy matching, existing mappings, and NLP.

    Args:
        :param str mdb_uri: MDB instance bolt uri
        :param str mdb_user: MDB instance username
        :param str mdb_pass: MDB instance password
        :param str base_model: model name (e.g., GDC)
        :param str comp_models: model names to compare to base model nodes/props
        :param str output_filepath: file path for output CSV
        :param float sim_threshold: threshold for similarty score results
        :param float num_nlp: number of nlp matches to return
            (nlp matches sorted by descending similarity score)

    Output:
        .csv file with base node/property handles and any matching node/property handles
    """
    filepath = Path(output_filepath)
    smdb = SearchableMDB(uri=mdb_uri, user=mdb_user, password=mdb_pass)
    tmdb = ToolsMDB(uri=mdb_uri, user=mdb_user, password=mdb_pass)

    # load spaCy NER model
    nlp = spacy.load("en_ner_bionlp13cg_md")

    # remove base model from comp_models if included for some reason
    comp_models = list(comp_model)
    if base_model in comp_models:
        comp_models.remove(base_model)

    base_nodes_props = []

    # gather all node and property handles for given base model
    for node in smdb.get_nodes_and_props_by_model(base_model):
        for prop in node["props"]:
            base_nodes_props.append(
                {
                    "model": base_model,
                    "node_handle": node["handle"],
                    "handle": prop["handle"],
                    "nanoid": prop["nanoid"],
                    "mapped_synonyms": [],
                    "fuzzy_matches": [],
                    "nlp_matches": [],
                }
            )

    comp_nodes_props = []

    # gather all node and property handles for given base model
    for model in comp_models:
        for node in smdb.get_nodes_and_props_by_model(model):
            for prop in node["props"]:
                comp_nodes_props.append(
                    {
                        "model": model,
                        "node_handle": node["handle"],
                        "handle": prop["handle"],
                        "nanoid": prop["nanoid"],
                    }
                )

    # keys=comp property nanoids; values=lists of corresponding node handles
    # used to get node handle for fuzzy props
    comp_nano_node_dict = {}
    for d in comp_nodes_props:
        if d["nanoid"] in comp_nano_node_dict:
            comp_nano_node_dict[d["nanoid"]].append(d["node_handle"])
            continue
        comp_nano_node_dict[d["nanoid"]] = [d["node_handle"]]

    for base_prop in base_nodes_props:
        # replace characters in property handle for fuzzy and nlp matching
        base_hdl_wc = base_prop["handle"].replace("_", "*")
        base_hdl_nlp = nlp(base_prop["handle"].replace("_", " ").lower())

        # get any existing synonyms of property
        base_synonyms = tmdb.get_property_synonyms_all(
            Property({"nanoid": base_prop["nanoid"]})
        )
        if base_synonyms:
            for synonym in base_synonyms:
                base_prop["mapped_synonyms"].append(
                    {
                        "nanoid": synonym.nanoid,
                        "model": synonym.model,
                        "handle": synonym.handle,
                    }
                )

        # fuzzy search on index for property handle matches
        fuzzy_match_list = _get_prop_fuzzy_matches(smdb, base_hdl_wc)
        if fuzzy_match_list:
            # add fuzzy match to fuzzy matches of base property
            for fuzzy_match in fuzzy_match_list:
                if (
                    fuzzy_match["model"] == base_model
                    or fuzzy_match in base_prop["fuzzy_matches"]
                ):
                    continue
                base_prop["fuzzy_matches"].append(fuzzy_match)

            # get list of mapped synonyms to fuzzy matches
            fuzzy_synonym_list = _get_fuzzy_synonyms(tmdb, fuzzy_match_list)

            # add fuzzy match synonyms to fuzzy matches of base property
            for fuzzy_synonym in fuzzy_synonym_list:
                if (
                    fuzzy_synonym["model"] == base_model
                    or fuzzy_synonym in base_prop["fuzzy_matches"]
                ):
                    continue
                base_prop["fuzzy_matches"].append(fuzzy_synonym)

        # add node handles to fuzzy matches
        for f_match in base_prop["fuzzy_matches"]:
            f_match["node_handles"] = comp_nano_node_dict[f_match["nanoid"]]

        # nlp
        nlp_match_list = []
        for comp_prop in comp_nodes_props:
            comp_hdl_nlp = nlp(comp_prop["handle"].replace("_", " ").lower())
            similarity = base_hdl_nlp.similarity(comp_hdl_nlp)
            if similarity >= sim_threshold:
                nlp_match_list.append(
                    {
                        "nanoid": comp_prop["nanoid"],
                        "model": comp_prop["model"],
                        "handle": comp_prop["handle"],
                        "node_handle": comp_prop["node_handle"],
                        "similarity": similarity,
                    }
                )
        # add top n nlp matches by similarity to nlp matches of base property
        base_prop["nlp_matches"] = sorted(
            nlp_match_list, key=lambda d: (d["similarity"], d["handle"]), reverse=True
        )[:num_nlp]

    df_cols = [f"{base_model}_node", f"{base_model}_prop"]
    for mod in comp_models:
        df_cols.extend([f"{mod}_fuzzy", f"{mod}_nlp"])
    df = pd.DataFrame(columns=df_cols)

    for base_prop in base_nodes_props:
        row = [base_prop["node_handle"], base_prop["handle"]]
        for model in comp_models:
            row.extend(
                [
                    _join_model_fuzzy_matches(base_prop, model),
                    _join_model_nlp_matches(base_prop, model),
                ]
            )
        df.loc[len(df.index)] = row

    df.to_csv(filepath, index=False)


if __name__ == "__main__":
    main()



"""
Script takes two MDF files representing different versions of the same
model and produces a Liquibase Changelog with the necessary changes to
an MDB in Neo4J to update the model from the old version to the new one.
"""

import configparser
import logging
from pathlib import Path
from typing import Dict, Generator, List, Optional, Tuple, Union

import click
from bento_mdf.diff import diff_models
from bento_mdf.mdf import MDF
from bento_meta.entity import Entity
from bento_meta.objects import Edge, Node, Property, Term
from bento_meta.util.cypher.clauses import (
    Delete,
    DetachDelete,
    Match,
    Merge,
    Remove,
    Set,
    Statement,
)
from bento_meta.util.cypher.entities import G, N, P, R, T, _plain_var
from liquichange.changelog import Changelog, Changeset, CypherChange

logger = logging.getLogger(__name__)

ADD_NODE = "add property graph node"
REMOVE_NODE = "remove property graph node"
ADD_RELATIONSHIP = "add property graph relationship"
REMOVE_RELATIONSHIP = "remove property graph relationship"
ADD_PROPERTY = "add property graph property"
REMOVE_PROPERTY = "remove property graph property"


def split_diff(
    diff: Dict[str, Dict[str, Optional[Union[List, Dict]]]], model_handle: str
) -> List[Tuple[str, Union[Entity, dict]]]:
    """splits model diff into segments that represent one change each"""
    diff_segments = []
    diff_order = [
        REMOVE_NODE,
        ADD_NODE,
        REMOVE_PROPERTY,
        ADD_PROPERTY,
        REMOVE_RELATIONSHIP,
        ADD_RELATIONSHIP,
    ]

    def add_node(entity, diff_segments):
        diff_segments.append((ADD_NODE, entity))

    def remove_node(entity, diff_segments):
        diff_segments.append((REMOVE_NODE, entity))

    def add_relationship(src, rel, dst, diff_segments):
        diff_segments.append((ADD_RELATIONSHIP, {"rel": rel, "src": src, "dst": dst}))

    def remove_relationship(src, rel, dst, diff_segments):
        diff_segments.append(
            (REMOVE_RELATIONSHIP, {"rel": rel, "src": src, "dst": dst})
        )

    def add_property(entity, prop_handle, prop_value, diff_segments):
        diff_segments.append(
            (
                ADD_PROPERTY,
                {
                    "entity": entity,
                    "prop_handle": prop_handle,
                    "prop_value": prop_value,
                },
            )
        )

    def remove_property(entity, prop_handle, prop_value, diff_segments):
        diff_segments.append(
            (
                REMOVE_PROPERTY,
                {
                    "entity": entity,
                    "prop_handle": prop_handle,
                    "prop_value": prop_value,
                },
            )
        )

    node_diff = diff.get("nodes")
    edge_diff = diff.get("edges")
    prop_diff = diff.get("props")

    if node_diff:
        if node_diff.get("a"):
            for node_hdl in node_diff.get("a"):
                remove_node(
                    Node({"handle": node_hdl, "model": model_handle}), diff_segments
                )
        if node_diff.get("b"):
            for node_hdl in node_diff.get("b"):
                add_node(
                    Node({"handle": node_hdl, "model": model_handle}), diff_segments
                )
        for node_hdl, change in node_diff.items():
            if node_hdl in {"a", "b"}:
                continue
            node_props = change["props"]
            node = Node({"handle": node_hdl, "model": model_handle})
            if node_props.get("a"):
                for prop_hdl in node_props.get("a"):
                    remove_relationship(
                        node,
                        "has_property",
                        Property({"handle": prop_hdl, "model": model_handle}),
                        diff_segments,
                    )
            if node_props.get("b"):
                for prop_hdl in node_props.get("b"):
                    add_relationship(
                        node,
                        "has_property",
                        Property({"handle": prop_hdl, "model": model_handle}),
                        diff_segments,
                    )
    if edge_diff:
        if edge_diff.get("a"):
            for edge_hdl, src_hdl, dst_hdl in edge_diff.get("a"):
                edge = Edge(
                    {
                        "handle": edge_hdl,
                        "src": Node({"handle": src_hdl, "model": model_handle}),
                        "dst": Node({"handle": dst_hdl, "model": model_handle}),
                    }
                )
                remove_node(edge, diff_segments)
        if edge_diff.get("b"):
            for edge_hdl, src_hdl, dst_hdl in edge_diff.get("b"):
                edge = Edge(
                    {
                        "handle": edge_hdl,
                        "src": Node({"handle": src_hdl, "model": model_handle}),
                        "dst": Node({"handle": dst_hdl, "model": model_handle}),
                    }
                )
                add_node(edge, diff_segments)
                add_relationship(edge, "has_src", edge.src, diff_segments)
                add_relationship(edge, "has_dst", edge.dst, diff_segments)
        for edge_tup, change_dict in edge_diff.items():
            if edge_tup in {"a", "b"}:
                continue
            edge = Edge(
                {
                    "handle": edge_tup[0],
                    "src": Node({"handle": edge_tup[1], "model": model_handle}),
                    "dst": Node({"handle": edge_tup[2], "model": model_handle}),
                }
            )
            for edge_attr, change in change_dict.items():
                if edge_attr == "props" and change.get("a"):
                    for prop_hdl in change.get("a"):
                        remove_relationship(
                            edge,
                            "has_property",
                            Property({"handle": prop_hdl, "model": model_handle}),
                            diff_segments,
                        )
                if edge_attr == "props" and change.get("b"):
                    for prop_hdl in change.get("b"):
                        add_relationship(
                            edge,
                            "has_property",
                            Property({"handle": prop_hdl, "model": model_handle}),
                            diff_segments,
                        )
                else:
                    if change.get("a"):
                        remove_property(
                            edge,
                            edge_attr,
                            change.get("a"),
                            diff_segments,
                        )
                    if change.get("b"):
                        add_property(
                            edge,
                            edge_attr,
                            change.get("b"),
                            diff_segments,
                        )
    if prop_diff:
        if prop_diff.get("a"):
            for prop_tuple in prop_diff.get("a"):
                # parent_hdls = prop_tuple[:-1]
                prop_hdl = prop_tuple[-1]
                prop = Property({"handle": prop_hdl, "model": model_handle})
                remove_node(prop, diff_segments)
        if prop_diff.get("b"):
            for prop_tuple in prop_diff.get("b"):
                # parent_hdls = prop_tuple[:-1]
                prop_hdl = prop_tuple[-1]
                prop = Property({"handle": prop_hdl, "model": model_handle})
                add_node(prop, diff_segments)
        for prop_tup, change_dict in prop_diff.items():
            if prop_tup in {"a", "b"}:
                continue
            prop = Property({"handle": prop_tup[1]})
            for prop_attr, change in change_dict.items():
                if prop_attr == "value_set" and change.get("a"):
                    value_set = change.get("a")
                    remove_relationship(
                        prop,
                        "has_value_set",
                        value_set,
                        diff_segments,
                    )
                if prop_attr == "value_set" and change.get("b"):
                    value_set = change.get("b")
                    add_node(value_set, diff_segments)
                    add_relationship(
                        prop,
                        "has_value_set",
                        value_set,
                        diff_segments,
                    )
                    for term in value_set.terms:
                        term_ent = Term({"value": term})
                        add_node(term_ent, diff_segments)
                        add_relationship(value_set, "has_term", term_ent, diff_segments)
                else:
                    if change.get("a"):
                        remove_property(
                            prop,
                            prop_attr,
                            change.get("a"),
                            diff_segments,
                        )
                    if change.get("b"):
                        add_property(
                            prop,
                            prop_attr,
                            change.get("b"),
                            diff_segments,
                        )
    return sorted(diff_segments, key=lambda x: diff_order.index(x[0]))


def convert_diff_segment_to_cypher_statement(
    diff_segment: Tuple[str, Union[Entity, dict]]
) -> Statement:
    """converts a diff segment to a changeset"""
    change, item = diff_segment
    if change == ADD_NODE:
        ent = N(label=item.get_label(), props=item.get_attr_dict())
        stmt = Statement(Merge(ent))
    elif change == REMOVE_NODE:
        ent = N(label=item.get_label(), props=item.get_attr_dict())
        if type(item) == Edge:
            src = N(label="node", props=item.src.get_attr_dict())
            dst = N(label="node", props=item.dst.get_attr_dict())
            src_trip = T(ent, R(Type="has_src"), src)
            dst_trip = T(ent, R(Type="has_dst"), dst)
            path = G(src_trip, dst_trip)
            match_clause = Match(path)
        else:
            match_clause = Match(ent)
        stmt = Statement(match_clause, DetachDelete(ent.var))
    elif change == ADD_RELATIONSHIP:
        rel = R(Type=item["rel"])
        src = N(label=item["src"].get_label(), props=item["src"].get_attr_dict())
        dst = N(label=item["dst"].get_label(), props=item["dst"].get_attr_dict())
        plain_trip = T(_plain_var(src), rel, _plain_var(dst))
        stmt = Statement(Match(src, dst), Merge(plain_trip))
    elif change == REMOVE_RELATIONSHIP:
        rel = R(Type=item["rel"])
        src = N(label=item["src"].get_label(), props=item["src"].get_attr_dict())
        dst = N(label=item["dst"].get_label(), props=item["dst"].get_attr_dict())
        trip = T(src, rel, dst)
        stmt = Statement(Match(trip), Delete(_plain_var(rel)))
    elif change == ADD_PROPERTY:
        prop = P(handle=item["prop_handle"], value=item["prop_value"])
        ent = N(label=item["entity"].get_label(), props=item["entity"].get_attr_dict())
        ent._add_props(prop)
        stmt = Statement(Match(ent), Set(ent.props[item["prop_handle"]]))
    elif change == REMOVE_PROPERTY:
        ent = N(label=item["entity"].get_label(), props=item["entity"].get_attr_dict())
        stmt = Statement(Match(ent), Remove(ent, prop=item["prop_handle"]))
    else:
        raise RuntimeError(f"invalid diff segment {diff_segment}")
    return stmt


def get_initial_changeset_id(config_file_path: str) -> int:
    """Gets initial changeset id from changelog config file"""
    config = configparser.ConfigParser()
    config.read(config_file_path)
    try:
        return config.getint(section="changelog", option="changeset_id")
    except (configparser.NoSectionError, configparser.NoOptionError) as error:
        print(error)
        raise


def changeset_id_generator(config_file_path: str) -> Generator[int, None, None]:
    """
    Iterator for changeset_id. Gets latest changeset id from changelog.ini
    and provides up-to-date changeset id as
    """
    i = get_initial_changeset_id(config_file_path)
    while True:
        yield i
        i += 1


def update_config_changeset_id(config_file_path: str, new_changeset_id: int) -> None:
    """Updates changelog config file with new changeset id"""
    config = configparser.ConfigParser()
    config.read(config_file_path)
    config.set(section="changelog", option="changeset_id", value=str(new_changeset_id))
    with open(file=config_file_path, mode="w", encoding="UTF-8") as config_file:
        config.write(fp=config_file)


def convert_diff_to_changelog(
    diff, model_handle: str, author: str, config_file_path: str
) -> Changelog:
    """converts diff beween two models to a changelog"""
    changeset_id = changeset_id_generator(config_file_path=config_file_path)
    changelog = Changelog()
    diff_segments = split_diff(diff, model_handle)

    for segment in diff_segments:
        cypher_stmt = convert_diff_segment_to_cypher_statement(segment)
        changelog.add_changeset(
            Changeset(
                id=str(next(changeset_id)),
                author=author,
                change_type=CypherChange(text=str(cypher_stmt)),
            )
        )

    update_config_changeset_id(
        config_file_path=config_file_path, new_changeset_id=next(changeset_id)
    )

    return changelog


@click.command()
@click.option(
    "--model_handle",
    required=True,
    type=str,
    prompt=True,
    help="CRDC Model Handle (e.g. 'GDC')",
)
@click.option(
    "--old_mdf_files",
    required=True,
    type=click.Path(exists=True),
    prompt=True,
    multiple=True,
    help="Older version of MDF file(s)",
)
@click.option(
    "--new_mdf_files",
    required=True,
    type=click.Path(exists=True),
    prompt=True,
    multiple=True,
    help="Newer version of MDF file(s)",
)
@click.option(
    "--output_file_path",
    required=True,
    type=click.Path(),
    prompt=True,
    help="File path for output changelog",
)
@click.option(
    "--config_file_path",
    required=True,
    type=click.Path(exists=True),
    help="Changeset config file path",
)
@click.option(
    "--author",
    required=True,
    type=str,
    help="author for changeset. default=MDB_ADMIN",
)
@click.option(
    "--_commit",
    required=False,
    type=str,
    help="commit string",
)
def main(
    model_handle: str,
    old_mdf_files: Union[str, List[str]],
    new_mdf_files: Union[str, List[str]],
    output_file_path: str,
    config_file_path: str,
    author: str,
    _commit: Optional[str],
) -> None:
    """
    get liquibase changelog from different versions of mdf files for a model
    """
    mdf_old = MDF(*old_mdf_files, handle=model_handle, _commit=_commit, raiseError=True)
    mdf_new = MDF(*new_mdf_files, handle=model_handle, _commit=_commit, raiseError=True)
    if not mdf_old.model or not mdf_new.model:
        raise RuntimeError("Error getting model from MDF")

    model_old = mdf_old.model
    model_new = mdf_new.model

    diff = diff_models(mdl_a=model_old, mdl_b=model_new)
    changelog = convert_diff_to_changelog(
        diff=diff,
        model_handle=model_handle,
        author=author,
        config_file_path=config_file_path,
    )

    path = Path(output_file_path)
    changelog.save_to_file(file_path=str(path), encoding="UTF-8")


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""
mdb.loaders: load models into an MDB instance consistently
"""

from tqdm import tqdm
from bento_meta.mdb import WriteableMDB
from bento_meta.util.cypher.entities import (
    N, R, P, T, G, _plain_var
    )
from bento_meta.util.cypher.clauses import (
    Match, Create, Merge, Set, OnMatchSet, OnCreateSet,
    Remove, Statement,
    )


def load_mdf(mdf, mdb, _commit=None):
    """Load an MDF object into an MDB instance"""
    load_model(mdf.model, mdb, _commit)


def load_model(model, mdb, _commit=None):
    """Load a model object into an MDB instance."""
    if not isinstance(mdb, WriteableMDB):
        raise RuntimeError("mdb object must be a WriteableMDB")
    cstmts = load_model_statements(model, _commit)
    for stmt in tqdm(cstmts):
        mdb.put_with_statement(str(stmt), stmt.params)


def load_model_statements(model, _commit=None):
    """
    Create Cypher statements from a model to load it de novo into an
    MDB instance.

    :param :class:`mdb.Model` model: Model instance for loading
    :param str _commit: 'Commit string' for marking entities in DB. If set, this will override
    _commit attributes already existing on Model entities.
    """
    cStatements = []
    cNodes = {}
    for nd in model.nodes:
        node = model.nodes[nd]
        cNode = _cEntity(node, model, _commit)
        cStatements.append(
            Statement(
                Merge(cNode),
                use_params=True
                )
            )
        cNodes[node.handle] = cNode
        cProps = []
        if node.tags:
            cStatements.extend(
                _tag_statements(node, cNode, _commit)
                )
        if node.props:
            cStatements.extend(
                _prop_statements(node, cNode, model, _commit)
                )
        if node.concept:
            cStatements.extend(
                _annotate_statements(node, cNode, _commit)
            )
    # node nodes and node-property nodes now exist
    # nodes are linked to properties
    for rl in model.edges:
        edge = model.edges[rl]
        cEdge = _cEntity(edge, model, _commit)
        cSrc = _cEntity(edge.src, model, _commit)
        cDst = _cEntity(edge.dst, model, _commit)

        if edge.multiplicity:
            cEdge._add_props({"multiplicity": edge.multiplicity})
        if edge.is_required:
            cEdge._add_props({"is_required": edge.is_required})
        # ensure uniqueness for merge
        cEdge._add_props({"__u":str(rl)})
        cStatements.extend([
            Statement(
                Create(cEdge),
                use_params=True
                ),
            Statement(
                Match(cEdge, cSrc, cDst),
                Merge(
                    R(Type="has_src").relate(_plain_var(cEdge), _plain_var(cSrc))
                    ),
                Merge(
                      R(Type="has_dst").relate(_plain_var(cEdge), _plain_var(cDst))
                ),
                use_params=True
                )])
        if edge.tags:
            cStatements.extend(
                _tag_statements(edge, cEdge, _commit)
                )
        if edge.props:
            cStatements.extend(
                _prop_statements(edge, cEdge, model, _commit)
                )
        if edge.concept:
            cStatements.extend(
                _annotate_statements(edge, cEdge, _commit)
                )
        cStatements.append(
            Statement(
                Match(cEdge),
                Remove(cEdge, prop="__u"),
                use_params=True
                )
            )
    # edge node and edge-property nodes now exist

    # now go through all properties that the model object knows about
    # - these should already have been created or merged, but
    # - if the property list on the model is missing any properties
    # - that nodes or edges have on themselves, then this indicates
    # - a bug/inconsistency  - and the property won't receive its
    # - value_set/term list in the DB in the following code.

    for pr in [x for x in model.props.values()
               if x.value_domain == 'value_set']:
        cValueSet = _cEntity(pr.value_set, model, _commit)
        cProp = _cEntity(pr, model, _commit)
        cStatements.extend([
            Statement(
                Merge(cValueSet),
                use_params=True
                ),
            Statement(
                Match(cProp, cValueSet),
                Merge(R(Type="has_value_set").
                      relate(_plain_var(cProp),
                             _plain_var(cValueSet))),
                use_params=True
                )
            ])
        if pr.concept:
            cStatements.extend(
                _annotate_statements(pr, cProp, _commit)
                )
        for tm in pr.terms.values():
            cTerm = _cEntity(tm, model, _commit)
            cStatements.extend([
                Statement(
                    Merge(cTerm),
                    use_params=True
                    ),
                Statement(
                    Match(cValueSet, cTerm),
                    Merge(R(Type="has_term").relate(_plain_var(cValueSet),
                                                    _plain_var(cTerm))),
                    use_params=True
                    )
                ])
    return cStatements


def _cEntity(ent, model, _commit):
    label = type(ent).__name__.lower()
    cEnt = None

    # translate labels
    if label == 'edge':
        label = 'relationship'
    if label == 'valueset':
        label = 'value_set'

    # special handling
    if label == 'term':
        cEnt = N(label=label,
                 props={"value": ent.value})
        if ent.origin_name:
            cEnt._add_props({"origin_name": ent.origin_name})
        if ent.origin_id:
            cEnt._add_props({"origin_id": ent.origin_id})
        if ent.origin_version:
            cEnt._add_props({"origin_version": ent.origin_version})
        if ent.origin_definition:
            cEnt._add_props({"origin_defintion": ent.origin_definition})
        if ent.handle:
            cEnt._add_props({"handle": ent.handle})
    elif label == 'value_set':
        cEnt = N(label='value_set',
                 props={"handle": ent.handle})
        if ent.url:
            cEnt._add_props({"url": ent.url})
    elif label == 'concept':
        cEnt = N(label='concept')
    elif label == 'tag':
        cEnt = N(label="tag",
                 props={"key": ent.key,
                        "value": ent.value})
        
    else:
        cEnt = N(label=label,
                 props={"handle": ent.handle,
                        "model": model.handle})
    # all ents
    if _commit:
        cEnt._add_props({"_commit": _commit})
    else:
        if ent._commit:
            cEnt._add_props({"_commit": ent._commit})
    if ent.nanoid:
        cEnt._add_props({"nanoid": ent.nanoid})
    if ent.desc:
        cEnt._add_props({"desc": ent.desc})
    return cEnt


def _tag_statements(ent, cEnt, _commit):
    stmts = []
    cTags = []
    if ent.tags:
        for t in ent.tags.values():
            cTag = _cEntity(t, None, _commit)
            cTags.append(cTag)
    for ct in cTags:
        stmts.append(
            Statement(
                Match(cEnt),
                Merge(R(Type="has_tag").relate(_plain_var(cEnt), ct)),
                use_params=True
                )
            )
    return stmts


def _prop_statements(ent, cEnt, model, _commit):
    stmts = []
    for p in ent.props.values():
        cProp = _cEntity(p, model, _commit)
        cProp._add_props({"value_domain": p.value_domain})
        stmts.extend([
            Statement(
                Merge(cProp),
                use_params=True
                ),
            Statement(
                Match(cEnt, cProp),
                Merge(R(Type="has_property").relate(
                    _plain_var(cEnt), _plain_var(cProp))),
                use_params=True
                ),
            ])
        if p.tags:
            stmts.extend(
                _tag_statements(p, cProp, _commit)
                )
    return stmts


def _annotate_statements(ent, cEnt, _commit):
    stmts = []
    if not ent.concept:
        return []
    cConcept = _cEntity(ent.concept, None, _commit)
    stmts = []
    for tm in ent.concept.terms.values():
        cTerm = _cEntity(tm, None, _commit)
        stmts.extend([
            Statement(
                Merge(cTerm),
                use_params=True
            ),
            Statement(
                Match(cEnt),
                Merge(R(Type="has_concept").relate(
                    _plain_var(cEnt), cConcept)),
                use_params=True
            ),
            Statement(
                Match(R(Type="has_concept").relate(
                    cEnt, cConcept), cTerm),
                Merge(R(Type="represents").relate(
                    _plain_var(cTerm), _plain_var(cConcept))),
                use_params=True
            ),
        ])
    return stmts



"""
ToolsMDB: subclass of 'WriteableMDB' to support interactions with the MDB.

EntityValidator: validates that entities have required attributes.
"""
import csv
import logging
from importlib.util import find_spec
from logging.config import fileConfig
from pathlib import Path
from subprocess import check_call
from sys import executable
from typing import Dict, Iterable, List, Optional, Set, Tuple, Type, Union

from bento_meta.entity import Entity
from bento_meta.mdb import make_nanoid, read_txn_data, read_txn_value
from bento_meta.mdb.writeable import WriteableMDB, write_txn
from bento_meta.objects import (
    Concept,
    Edge,
    Node,
    Predicate,
    Property,
    Tag,
    Term,
    ValueSet,
)
from bento_meta.util.cypher.clauses import (
    As,
    Collect,
    DetachDelete,
    Match,
    Merge,
    OptionalMatch,
    Return,
    Statement,
    With,
)
from bento_meta.util.cypher.entities import N0, R0, G, N, P, R, T, _plain_var
from bento_meta.util.cypher.functions import count

# logging stuff
log_ini_path = Path(__file__).parents[2].joinpath("logs/log.ini")
log_file_path = Path(__file__).parents[2].joinpath(f"logs/{__name__}.log")
fileConfig(log_ini_path, defaults={"logfilename": log_file_path.as_posix()})
logger = logging.getLogger(__name__)


class ToolsMDB(WriteableMDB):
    """Adds mdb-tools to WriteableMDB"""

    def __init__(self, uri, user, password):
        WriteableMDB.__init__(self, uri, user, password)

    class EntityNotUniqueError(Exception):
        """Raised when an entity's attributes identify more than 1 property graph node in an MDB"""

    class EntityNotFoundError(Exception):
        """Raised when an entity's attributes fail to identify a property graph node in an MDB"""

    class PatternNotUniqueError(Exception):
        """
        Raised when a match pattern's attributes identify more than 1
        property graph triple or set of overlapping triples in an MDB
        """

    class PatternNotFoundError(Exception):
        """
        Raised when a match pattern's attributes fail to identify
        a property graph triple or set of overlapping triples in an MDB
        """

    @read_txn_value
    def _get_entity_count(self, entity: Entity):
        """
        Returns count of given entity found in an MDB.

        If count = 0, entity with given attributes not found in MDB.
        If count = 1, entity with given attributes is unique in MDB
        If count > 1, more attributes needed to uniquely id entity in MDB.
        """
        ent = N(label=entity.get_label(), props=entity.get_attr_dict())

        stmt = Statement(
            Match(ent),
            Return(count(ent.var)),
            As("entity_count"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "entity_count")

    @read_txn_value
    def _get_pattern_count(self, pattern: Union[T, G]):
        """
        Returns count of given match pattern, which could be a triple like:
        (n)-[r]->(m), or a set of overlapping triples (Path) found in MDB.

        If count = 0, pattern with given attributes not found in MDB.
        If count = 1, pattern with given attributes is unique in MDB
        If count > 1, more attributes needed to uniquely id pattern in MDB.
        """
        stmt = Statement(
            Match(pattern), Return(count("*")), As("pattern_count"), use_params=True
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "pattern_count")

    def validate_entity_unique(self, entity: Entity) -> None:
        """
        Validates that the given entity occurs once (& only once) in an MDB

        Raises EntityNotUniqueError if entity attributes match multiple property
            graph nodes in the MDB.

        Raises EntityNotFoundError if entity attributes don't match any in the MDB.

        Note: doesn't validate the entity itself because not all entity attibutes
            necessarily required to locate an entity in the MDB.
            (e.g. handle and model OR nanoid alone can identify a node)
        """

        ent_count = int(self._get_entity_count(entity)[0])
        if ent_count > 1:
            logger.error(str(self.EntityNotUniqueError))
            raise self.EntityNotUniqueError
        if ent_count < 1:
            logger.error(str(self.EntityNotFoundError))
            raise self.EntityNotFoundError

    def validate_entities_unique(self, entities: Iterable[Entity]) -> None:
        """Runs self.validate_entity_unique() over multiple entities"""
        for entity in entities:
            self.validate_entity_unique(entity)

    def validate_pattern_unique(self, pattern: Union[T, G]) -> None:
        """
        Validates that the given match pattern occurs once (& only once) in an MDB

        Raises PatternNotUniqueError if entity attributes match multiple property
        graph nodes in the MDB.

        Raises PatternNotFoundError if entity attributes don't match any in the MDB.
        """
        pattern_count = int(self._get_pattern_count(pattern)[0])
        if pattern_count > 1:
            logger.error(
                str(
                    self.PatternNotUniqueError(
                        f"Pattern: {pattern.pattern()} not unique."
                    )
                )
            )
            raise self.PatternNotUniqueError(
                f"Pattern: {pattern.pattern()} not unique."
            )
        if pattern_count < 1:
            logger.error(
                str(
                    self.PatternNotFoundError(
                        f"Pattern: {pattern.pattern()} not found."
                    )
                )
            )
            raise self.PatternNotFoundError(f"Pattern: {pattern.pattern()} not found.")

    @write_txn
    def remove_entity_from_mdb(self, entity: Entity):
        """
        Remove given Entity node from the database.

        Accepts the following bento-meta Entities:
            Concept, Node, Predicate, Property, Edge, Term
        """
        self.validate_entity_unique(entity)

        ent_label = entity.get_label()
        ent_attrs = entity.get_attr_dict()

        ent = N(label=ent_label, props=ent_attrs)

        stmt = Statement(Match(ent), DetachDelete(ent.var), use_params=True)

        qry = str(stmt)
        parms = stmt.params

        logging.info(f"Removing {ent_label} entity with with properties: {ent_attrs}")
        return (qry, parms)

    @write_txn
    def add_entity_to_mdb(
        self,
        entity: Entity,
        _commit=None,
    ):
        """Adds given Entity node to MDB instance"""
        EntityValidator.validate_entity(entity)
        EntityValidator.validate_entity_has_attribute(entity, "nanoid")

        if _commit:
            entity._commit = _commit

        ent_label = entity.get_label()
        ent_attrs = entity.get_attr_dict()

        ent = N(label=ent_label, props=ent_attrs)

        if isinstance(entity, Edge):
            src = N(label="node", props=entity.src.get_attr_dict())
            dst = N(label="node", props=entity.dst.get_attr_dict())
            src_rel = R(Type="has_src")
            dst_rel = R(Type="has_dst")
            src_trip = src_rel.relate(_plain_var(ent), _plain_var(src))
            dst_trip = dst_rel.relate(_plain_var(ent), _plain_var(dst))
            stmt = Statement(
                Merge(ent),
                Merge(src),
                Merge(dst),
                Merge(src_trip),
                Merge(dst_trip),
                use_params=True,
            )
        else:
            stmt = Statement(Merge(ent), use_params=True)

        qry = str(stmt)
        parms = stmt.params

        logging.info(
            f"Merging new {ent_label} node with properties: {ent_attrs} into the MDB"
        )

        return (qry, parms)

    @read_txn_value
    def get_concept_nanoids_linked_to_entity(
        self, entity: Entity, mapping_source: Optional[str] = None
    ):
        """
        Returns list of concept nanoids linked to given entity by
        "represents" or "has_concept" relationships tagged with
        the given mapping source.
        """
        self.validate_entity_unique(entity)

        ent = N(label=entity.get_label(), props=entity.get_attr_dict())
        concept = N(label="concept")
        if isinstance(entity, Term):
            rel = R(Type="represents")
        else:
            rel = R(Type="has_concept")
        ent_trip = rel.relate(ent, concept)

        if mapping_source is not None:
            tag = N(
                label="tag",
                props={"key": "mapping_source", "value": mapping_source},
            )
            tag_trip = T(concept, R(Type="has_tag"), tag)
            path = G(ent_trip, tag_trip)
        else:
            path = ent_trip

        stmt = Statement(
            Match(path),
            Return(f"{concept.var}.nanoid"),
            As("concept_nanoids"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params
        logging.debug(f"{qry=}; {parms=}")

        return (qry, parms, "concept_nanoids")

    @write_txn
    def add_relationship_to_mdb(
        self,
        relationship_type: str,
        src_entity: Entity,
        dst_entity: Entity,
        _commit: str = "",
    ):
        """
        Adds relationship between given entities in MDB.
        """
        self.validate_entities_unique([src_entity, dst_entity])

        rel = R(Type=relationship_type)

        src = N(label=src_entity.get_label(), props=src_entity.get_attr_dict())
        dst = N(label=dst_entity.get_label(), props=dst_entity.get_attr_dict())
        trip = rel.relate(src, dst)
        plain_trip = rel.relate(_plain_var(src), _plain_var(dst))

        try:
            # check for existance shouldn't include _commit?
            self.validate_pattern_unique(trip)
        except self.PatternNotFoundError:
            # if triple doesn't already exist, add _commit? this means that
            # if triple exists w/ different _commit, merge shouldn't add anything?
            rel.props["_commit"] = P(handle="_commit", value=_commit)

        stmt = Statement(Match(src, dst), Merge(plain_trip), use_params=True)

        qry = str(stmt)
        parms = stmt.params

        print(stmt)

        logging.info(
            f"Adding {relationship_type} relationship between src {src.label} with "
            f"properties: {src_entity.get_attr_dict()} to dst {dst.label} "
            f"with properties: {dst_entity.get_attr_dict()}"
        )
        return (qry, parms)

    def link_synonyms(
        self,
        entity_1: Entity,
        entity_2: Entity,
        mapping_source: str,
        _commit: str = "",
    ) -> None:
        """
        Link two synonymous entities in the MDB via a Concept node.

        This function takes two synonymous Entities (as determined by user/SME) as
        bento-meta objects connects them to a Concept node via a 'represents' relationship.

        Entities must both exist in the MDB instance and given entity attributes must
        uniquely identify property graph nodes in the MDB.

        If _commit is set (to a string), the _commit property of any node created is set to
        this value.
        """
        self.validate_entities_unique([entity_1, entity_2])

        ent_1_concepts = self.get_concept_nanoids_linked_to_entity(
            entity=entity_1, mapping_source=mapping_source
        )
        ent_2_concepts = self.get_concept_nanoids_linked_to_entity(
            entity=entity_2, mapping_source=mapping_source
        )
        shared_concepts = list(set(ent_1_concepts).intersection(ent_2_concepts))

        # has concept been tagged by this mapping src before
        if shared_concepts:
            logging.warning(
                f"This mapping has already been added by this source via "
                f"Concept with nanoid: {shared_concepts[0]}"
            )
            return

        # one of the entities has a concept created by the given mapping source
        if ent_1_concepts:
            logging.info(f"Using existing concept with nanoid {ent_1_concepts[0]}")
            concept = Concept({"nanoid": ent_1_concepts[0]})
        elif ent_2_concepts:
            logging.info(f"Using existing concept with nanoid {ent_2_concepts[0]}")
            concept = Concept({"nanoid": ent_2_concepts[0]})
        else:
            concept = Concept({"nanoid": make_nanoid()})
            self.add_entity_to_mdb(concept, _commit=_commit)
            self.add_tag_to_mdb_entity(
                tag=Tag(
                    {
                        "key": "mapping_source",
                        "value": mapping_source,
                        "nanoid": make_nanoid(),
                    }
                ),
                entity=concept,
            )

        # create specified relationship between each entity and a concept
        rel_type_1 = "represents" if isinstance(entity_1, Term) else "has_concept"
        rel_type_2 = "represents" if isinstance(entity_2, Term) else "has_concept"

        self.add_relationship_to_mdb(
            relationship_type=rel_type_1,
            src_entity=entity_1,
            dst_entity=concept,
            _commit=_commit,
        )
        self.add_relationship_to_mdb(
            relationship_type=rel_type_2,
            src_entity=entity_2,
            dst_entity=concept,
            _commit=_commit,
        )

    @read_txn_value
    def get_entity_nanoid(self, entity: Entity):
        """
        Takes a unique entity in the MDB and returns its nanoid.
        """
        self.validate_entity_unique(entity)

        ent = N(label=entity.get_label(), props=entity.get_attr_dict())

        if isinstance(entity, Edge):
            src = N(label="node", props=entity.src.get_attr_dict())
            dst = N(label="node", props=entity.dst.get_attr_dict())
            src_rel = R(Type="has_src")
            dst_rel = R(Type="has_dst")
            src_trip = src_rel.relate(ent, src)
            dst_trip = dst_rel.relate(ent, dst)
            path = G(src_trip, dst_trip)
            match_clause = Match(path)
        else:
            match_clause = Match(ent)

        stmt = Statement(
            match_clause,
            Return(f"{ent.var}.nanoid"),
            As("entity_nanoid"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "entity_nanoid")

    def get_or_make_entity_nanoid(self, entity: Entity) -> str:
        """Obtains existing entity's nanoid or creates one for new entity."""
        try:
            return self.get_entity_nanoid(entity)[0]
        except self.EntityNotFoundError:
            nanoid = make_nanoid()
            return nanoid

    @read_txn_value
    def get_term_nanoids(self, concept: Concept, mapping_source: str = ""):
        """Returns list of term nanoids representing given concept"""
        self.validate_entity_unique(concept)

        ent = N(label="concept", props=concept.get_attr_dict())
        term = N(label="term")
        ent_trip = T(term, R(Type="represents"), ent)

        if mapping_source:
            tag_trip = T(
                concept,
                R(Type="has_tag"),
                N(
                    label="tag",
                    props={"key": "mapping_source", "value": mapping_source},
                ),
            )
            path = G(ent_trip, tag_trip)
        else:
            path = ent_trip

        stmt = Statement(
            Match(path),
            Return(f"{term.var}.nanoid"),
            As("term_nanoids"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "term_nanoids")

    @read_txn_value
    def get_predicate_nanoids(self, concept: Concept, mapping_source: str = ""):
        """Returns list of predicate nanoids with relationship to given concept"""
        self.validate_entity_unique(concept)

        ent = N(label="concept", props=concept.get_attr_dict())
        predicate = N(label="predicate")
        ent_trip = T(predicate, R0(), ent)

        if mapping_source:
            tag = N(
                label="tag",
                props={"key": "mapping_source", "value": mapping_source},
            )
            tag_trip = T(ent, R(Type="has_tag"), tag)
            path = G(ent_trip, tag_trip)
        else:
            path = ent_trip

        stmt = Statement(
            Match(path),
            Return(f"{predicate.var}.nanoid"),
            As("predicate_nanoids"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "predicate_nanoids")

    @read_txn_value
    def get_relationship_between_entities(self, src_entity: Entity, dst_entity: Entity):
        """Returns relationship type between given entities with (src)-[:rel_type]->(dst)"""
        self.validate_entities_unique([src_entity, dst_entity])

        trip = T(
            N(label=src_entity.get_label(), props=src_entity.get_attr_dict()),
            R(),
            N(label=dst_entity.get_label(), props=dst_entity.get_attr_dict()),
        )

        stmt = Statement(
            Match(trip),
            Return(f"TYPE({trip._edge.var})"),
            As("relationship_type"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "relationship_type")

    def link_concepts_via_predicate(
        self,
        subject_concept: Concept,
        object_concept: Concept,
        predicate: Predicate,
        _commit="",
    ) -> None:
        """
        Links two synonymous Concepts via a Predicate

        This function takes two synonymous Concepts as objects and links
        them via a Predicate node and has_subject and has_object relationships.
        """
        self.validate_entities_unique([subject_concept, object_concept])

        predicate.subject = subject_concept
        predicate.object = object_concept

        predicate.nanoid = self.get_or_make_entity_nanoid(predicate)
        self.add_entity_to_mdb(predicate, _commit=_commit)

        self.add_relationship_to_mdb(
            relationship_type="has_subject",
            src_entity=predicate,
            dst_entity=predicate.subject,
            _commit=_commit,
        )
        self.add_relationship_to_mdb(
            relationship_type="has_object",
            src_entity=predicate,
            dst_entity=predicate.object,
            _commit=_commit,
        )

    def merge_two_concepts(
        self,
        concept_1: Concept,
        concept_2: Concept,
        mapping_source: str = "",
        _commit="",
    ) -> None:
        """
        Combine two synonymous Concepts into a single Concept.

        This function takes two synonymous Concept as bento-meta objects and
        merges them into a single Concept along with any connected Terms and Predicates.
        """
        self.validate_entities_unique([concept_1, concept_2])

        # get list of terms connected to concept 2
        c2_term_nanoids = self.get_term_nanoids(concept_2, mapping_source)
        c2_terms: List[Term] = []
        for nanoid in c2_term_nanoids:
            c2_terms.append(Term({"nanoid": nanoid}))

        # get list of predicates connected to concept 2
        c2_predicate_nanoids = self.get_predicate_nanoids(concept_2, mapping_source)
        c2_predicates_with_rel = []
        for nanoid in c2_predicate_nanoids:
            predicate = Predicate({"nanoid": nanoid})
            predicate_rel = self.get_relationship_between_entities(
                src_entity=predicate, dst_entity=concept_2
            )[0]
            c2_predicates_with_rel.append((predicate, predicate_rel))

        # delete concept 2
        self.remove_entity_from_mdb(concept_2)

        # connect terms from deleted (c2) to remaining concept (c1)
        for term in c2_terms:
            self.add_relationship_to_mdb(
                relationship_type="represents",
                src_entity=term,
                dst_entity=concept_1,
                _commit=_commit,
            )

        # connect predicates from deleted (c2) to remaining concept (c1)
        for predicate, rel in c2_predicates_with_rel:
            self.add_relationship_to_mdb(
                relationship_type=rel,
                src_entity=predicate,
                dst_entity=concept_1,
                _commit=_commit,
            )

    @read_txn_data
    def _get_all_terms(self):
        """Returns list of all terms in an MDB."""
        term = N(label="term")

        stmt = Statement(Match(term), Return(term.var))

        qry = str(stmt)
        parms = {}

        print(qry)

        return (qry, parms)

    def get_potential_term_synonyms(
        self, term: Term, threshhold: float = 0.8
    ) -> List[dict]:
        """
        Returns list of dicts representing potential Term nodes synonymous to given Term
        in an MDB
        """
        self.validate_entity_unique(term)

        nlp = _get_nlp_model()

        all_terms_result = self._get_all_terms()
        all_terms = [list(item.values())[0] for item in all_terms_result]

        # get likely synonyms
        synonyms = []
        for term_attr_dict in all_terms:
            # calculate similarity between each Term and input Term
            term_1 = nlp(term.value)
            term_2 = nlp(term_attr_dict["value"])
            similarity_score = term_1.similarity(term_2)
            # if similarity threshold met, add to list of potential synonyms
            if similarity_score >= threshhold:
                synonym = {
                    "value": term_attr_dict["value"],
                    "origin_name": term_attr_dict["origin_name"],
                    "nanoid": term_attr_dict["nanoid"],
                    "similarity": similarity_score,
                    "valid_synonym": 0,  # mark 1 if synonym when uploading later
                }
                synonyms.append(synonym)
        synonyms_sorted = sorted(synonyms, key=lambda d: d["similarity"], reverse=True)
        return synonyms_sorted

    def potential_synonyms_to_csv(
        self, input_data: List[dict], output_path: str
    ) -> None:
        """Given a list of synonymous Terms as dicts, outputs to CSV file at given output path"""
        with open(output_path, "w", encoding="utf8", newline="") as output_file:
            dict_writer = csv.DictWriter(output_file, fieldnames=input_data[0].keys())
            dict_writer.writeheader()
            dict_writer.writerows(input_data)

    def link_term_synonyms_csv(
        self, term: Term, csv_path: str, mapping_source: str, _commit: str = ""
    ) -> None:
        """Given a CSV of syonymous Terms, links each via a Concept node to given Term"""
        with open(csv_path, encoding="UTF-8") as csvfile:
            synonym_reader = csv.reader(csvfile)
            for line in synonym_reader:
                if line[3] == "1":
                    synonym = Term()
                    synonym.value = line[0]
                    synonym.origin_name = line[1]
                    self.link_synonyms(
                        entity_1=term,
                        entity_2=synonym,
                        mapping_source=mapping_source,
                        _commit=_commit,
                    )

    @read_txn_data
    def get_property_synonyms_direct(self, entity: Property, mapping_source: str = ""):
        """
        Returns list of properties linked by concept to given property
        """
        self.validate_entity_unique(entity)

        ent = N(label="property", props=entity.get_attr_dict())
        prop = N(label="property")
        concept = N(label="concept")
        ent_trip_1 = T(ent, R(Type="has_concept"), concept)
        ent_trip_2 = T(prop, R(Type="has_concept"), concept)

        if mapping_source:
            tag = N(
                label="tag",
                props={"key": "mapping_source", "value": mapping_source},
            )
            tag_trip = T(concept, R(Type="has_tag"), tag)
            path = G(ent_trip_1, ent_trip_2, tag_trip)
        else:
            path = G(ent_trip_1, ent_trip_2)

        stmt = Statement(
            Match(path),
            With(f"{Collect(_plain_var(prop).pattern())}"),
            As("synonyms"),
            Return("synonyms"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms)

    def _get_property_synonyms_direct_as_list(self, entity: Property) -> List[Property]:
        """
        converts results of read_txn_data-wrapped function
        with one item to a simple list of bento_meta.objects.Property entities
        """
        data = self.get_property_synonyms_direct(entity)
        return [Property(s) for s in data[0]["synonyms"]]

    def get_property_synonyms_all(self, entity: Property) -> List[Property]:
        """
        Returns list of properties linked by concept to given property
        or to synonym of given property (and so on)
        """
        self.validate_entity_unique(entity)
        all_synonyms = []
        queue = [entity]
        visited = {entity.nanoid}

        while queue:
            current = queue.pop(0)
            direct_synonyms = self._get_property_synonyms_direct_as_list(current)
            for synonym in direct_synonyms:
                if synonym.nanoid not in visited:
                    visited.add(synonym.nanoid)
                    queue.append(synonym)
                    all_synonyms.append(synonym)

        return all_synonyms

    @read_txn_data
    def _get_property_parents_data(self, entity: Property):
        """get list of nodes/edges connected to given property
        via the "has_property" relationship"""
        self.validate_entity_unique(entity)
        p_attrs = entity.get_attr_dict()
        child_prop = N(label="property", props=p_attrs)
        parent_node = N(label="node")
        parent_edge = N(label="relationship")
        rel = R(Type="has_property", _dir="_left")
        trip1 = rel.relate(child_prop, N0())
        trip2 = rel.relate(_plain_var(child_prop), parent_node)
        trip3 = rel.relate(_plain_var(child_prop), parent_edge)

        stmt = Statement(
            Match(trip1),
            OptionalMatch(trip2),
            OptionalMatch(trip3),
            With(),
            Collect(_plain_var(parent_node).pattern()),
            As("nodes,"),
            Collect(_plain_var(parent_edge).pattern()),
            As("edges"),
            Return("nodes, edges"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms)

    def get_property_parents(self, entity: Property) -> List[Union[Node, Edge]]:
        """
        returns results of _get_property_parents_data as a list of
        bento_meta Nodes or Edges
        """
        self.validate_entity_unique(entity)

        data = self._get_property_parents_data(entity)
        node_parents = [Node(p) for p in data[0]["nodes"]]
        edge_parents = [Edge(p) for p in data[0]["edges"]]
        return node_parents + edge_parents

    def add_tag_to_mdb_entity(self, tag: Tag, entity: Entity) -> None:
        """Adds a tag to an existing entity in an MDB."""
        self.validate_entity_unique(entity)
        self.add_entity_to_mdb(tag)
        self.add_relationship_to_mdb(
            relationship_type="has_tag", src_entity=entity, dst_entity=tag
        )


class EntityValidator:
    """Entity validator that validate entities have all required attributes"""

    required_attrs_by_entity_type: Dict[Type[Entity], List[str]] = {
        Node: ["handle", "model"],
        Edge: ["handle", "model", "src", "dst"],
        Property: ["handle", "model"],
        Term: ["origin_name", "value"],  # add? "origin_id", "origin_version"
        Concept: ["nanoid"],
        Predicate: ["handle", "subject", "object"],
        Tag: ["key", "value"],
        ValueSet: ["handle"],
    }

    valid_attrs: Dict[Tuple[Type[Entity], str], Set[str]] = {
        (Predicate, "handle"): {
            "exactMatch",
            "closeMatch",
            "broader",
            "narrower",
            "related",
        }
    }

    class MissingAttributeError(Exception):
        """Raised when an entity doesn't have the attributes required for unique identification"""

    class InvalidAttributeError(Exception):
        """Raised when an entity attribute is invalid"""

    @staticmethod
    def validate_entity_has_attribute(entity: Entity, attr_name: str) -> None:
        """Validates the presence of an entity's attribute"""
        if not getattr(entity, attr_name):
            raise EntityValidator.MissingAttributeError(
                f"{entity.__class__.__name__} needs a {attr_name} attribute"
            )

    @staticmethod
    def _validate_entity_attribute(entity_type: Type[Entity], attr_name: str) -> None:
        """Checks that an entity attribute is in a set of valid attributes"""
        valid_attrs = EntityValidator.valid_attrs.get((entity_type, attr_name))

        if valid_attrs and getattr(entity_type, attr_name) not in valid_attrs:
            raise EntityValidator.InvalidAttributeError(
                f"{entity_type.__name__} {attr_name} must be one of: "
                f"{', '.join(list(valid_attrs))}"
            )

    @staticmethod
    def validate_entity(entity: Entity) -> None:
        """
        Checks if entity has all attributes required by MDB for its type, and
        that all of those attributes are valid themselves if they are entities or
        if they have a fixed set of possible values.

        Verifies that bento-meta entity has all necesssary attributes before
        it is added to an MDB instance.

        If looking for a unique identifier for the entity (i.e. nanoid), ensures
        entity has all the required attributes for unique identification.
        """
        entity_type = type(entity)
        required_attrs = EntityValidator.required_attrs_by_entity_type.get(entity_type)

        if not required_attrs:
            raise ValueError(f"Entity type {entity_type.__name__} not supported")

        for attr_name in required_attrs:
            try:
                # validate attr that is an entity (e.g. Edge.src)
                if isinstance(getattr(entity, attr_name), Entity):
                    EntityValidator.validate_entity(getattr(entity, attr_name))
                # check entity has attr if required
                EntityValidator.validate_entity_has_attribute(entity, attr_name)
                # check attr with fixed set of possible values is valid
                EntityValidator._validate_entity_attribute(entity_type, attr_name)
            except (
                EntityValidator.MissingAttributeError,
                EntityValidator.InvalidAttributeError,
            ) as error:
                logger.error(str(error))
                raise


def _get_nlp_model():
    """Installs and imports spacy & scispacy nlp model if any not already installed"""
    model_name = "en_ner_bionlp13cg_md"
    model_url = "https://s3-us-west-2.amazonaws.com/ai2-s2-scispacy/releases/v0.5.1/en_ner_bionlp13cg_md-0.5.1.tar.gz"

    try:
        if not find_spec("spacy"):  # ensure spacy is installed
            check_call([executable, "-m", "pip", "install", "spacy"])
        import spacy

        if not find_spec(model_name):  # ensure model is installed
            check_call([executable, "-m", "pip", "install", model_url])
        nlp = spacy.load(model_url)
        return nlp
    except Exception as error:
        logger.error(str(error))
        raise



"""converts mapping MDF to changelog"""
import configparser
from ast import literal_eval
from pathlib import Path
from typing import Dict, Generator, List, Optional, Union

import click
import yaml
from bento_meta.util.cypher.clauses import (
    Case,
    Create,
    ForEach,
    Match,
    Merge,
    OptionalMatch,
    Statement,
    When,
    With,
)
from bento_meta.util.cypher.entities import G, N, R, T, _anon, _plain_var
from liquichange.changelog import Changelog, Changeset, CypherChange


def get_initial_changeset_id(config_file_path: str) -> int:
    """Gets initial changeset id from changelog config file"""
    config = configparser.ConfigParser()
    config.read(config_file_path)
    try:
        return config.getint(section="changelog", option="changeset_id")
    except (configparser.NoSectionError, configparser.NoOptionError) as error:
        print(error)
        raise


def changeset_id_generator(config_file_path: str) -> Generator[int, None, None]:
    """
    Iterator for changeset_id. Gets latest changeset id from changelog.ini
    and provides up-to-date changeset id as
    """
    i = get_initial_changeset_id(config_file_path)
    while True:
        yield i
        i += 1


def update_config_changeset_id(config_file_path: str, new_changeset_id: int) -> None:
    """Updates changelog config file with new changeset id"""
    config = configparser.ConfigParser()
    config.read(config_file_path)
    config.set(section="changelog", option="changeset_id", value=str(new_changeset_id))
    with open(file=config_file_path, mode="w", encoding="UTF-8") as config_file:
        config.write(fp=config_file)


def load_yaml(
    file: str,
) -> Dict[
    str,
    Union[
        str,
        List[str],
        Dict[
            str,
            Dict[
                str,
                Dict[str, Dict[str, List[Dict[str, Dict[str, Union[str, bool]]]]]],
            ],
        ],
    ],
]:
    """loads mapping Dict from yaml"""
    with open(file, "r", encoding="UTF8") as stream:
        try:
            data = yaml.safe_load(stream)
            return data
        except yaml.YAMLError as exc:
            print(exc)
            raise exc


def get_mapping_source(mapping_dict: Dict) -> str:
    """returns source of cross-model mappings"""
    return mapping_dict.get("Source", "")


def get_target_models(mapping_dict: Dict) -> Dict[str, str]:
    """returns target models"""
    return mapping_dict.get("Models", dict())


def get_parents_as_list(parent_str: str) -> List[str]:
    """Returns parent nodes as a list."""
    if not parent_str or not isinstance(parent_str, str):
        raise ValueError("Input must be a non-empty string.")
    if parent_str.startswith("[") and parent_str.endswith("]"):
        return literal_eval(parent_str)
    if "." in parent_str:
        return parent_str.split(".")
    return [parent_str]


def generate_mapping_cypher(
    src_ent: N,
    dst_ent: N,
    src_parent: N,
    dst_parent: N,
    parent_child_rel: R,
    mapping_source: str,
    _commit: Optional[str] = None,
) -> Statement:
    """
    Matches existing src_ent & dst_ent with given parents/ancestors and
    generates cypher query to link synonymous entities via a Concept node.

    Adds Tag node {mapping_source: src_model} to generated concept.

    _commit will be added to newly created nodes
    """
    src_triple = T(src_parent, _anon(parent_child_rel), src_ent)
    dst_triple = T(dst_parent, _anon(parent_child_rel), dst_ent)
    src_concept = N(label="concept")
    dst_concept = N(label="concept")
    src_concept_path = G(
        T(_plain_var(src_ent), R(Type="has_concept"), src_concept),
        T(
            _plain_var(src_concept),
            R(Type="has_tag"),
            N(label="tag", props={"key": "mapping_source", "value": mapping_source}),
        ),
    )
    dst_concept_path = G(
        T(_plain_var(dst_ent), R(Type="has_concept"), dst_concept),
        T(
            _plain_var(dst_concept),
            R(Type="has_tag"),
            N(label="tag", props={"key": "mapping_source", "value": mapping_source}),
        ),
    )
    new_concept = N(label="concept", props={"_commit": _commit})
    new_tag = N(
        label="tag",
        props={"key": "mapping_source", "value": mapping_source, "_commit": _commit},
    )
    return Statement(
        # find src & dst entities using parent triples
        Match(src_triple, dst_triple),
        # check if src or dst ents have a concept tagged by the mapping src
        OptionalMatch(src_concept_path),
        OptionalMatch(dst_concept_path),
        With(src_ent.var, dst_ent.var, src_concept.var, dst_concept.var),
        # src ent has existing concept tagged by mapping src, link dst ent to it
        ForEach(),
        f"(_ IN {Case()}{When(f'{src_concept.var} IS NOT NULL', f'{dst_concept.var} IS NULL')} ",
        "THEN [1] ELSE [] END |",
        Merge(T(_plain_var(dst_ent), R(Type="has_concept"), _plain_var(src_concept))),
        ")",
        # dst ent has existing concept tagged by mapping src, link src ent to it
        ForEach(),
        f"(_ IN {Case()}{When(f'{src_concept.var} IS NULL', f'{dst_concept.var} IS NOT NULL')} ",
        "THEN [1] ELSE [] END |",
        Merge(T(_plain_var(src_ent), R(Type="has_concept"), _plain_var(dst_concept))),
        ")",
        # neither ent has a concept, create a new one, tag it, link ents to it
        ForEach(),
        f"(_ IN {Case()}{When(f'{src_concept.var} IS NULL', f'{dst_concept.var} IS NULL')} ",
        "THEN [1] ELSE [] END |",
        Create(T(new_concept, R(Type="has_tag"), new_tag)),
        Create(
            T(
                _plain_var(src_ent),
                R(Type="has_concept", props={"_commit": _commit}),
                _plain_var(new_concept),
            )
        ),
        Create(
            T(
                _plain_var(dst_ent),
                R(Type="has_concept", props={"_commit": _commit}),
                _plain_var(new_concept),
            )
        ),
        ")",
    )


def process_props(mapping_dict: Dict, _commit: Optional[str] = None) -> List[Statement]:
    """processes mappings between model Property entities"""
    cypher_stmts = []
    src_model = get_mapping_source(mapping_dict)
    prop_map = mapping_dict.get("Props")
    if not prop_map:
        raise RuntimeError("error loading 'Props'")
    for src_parent, src_prop_dict in prop_map.items():
        for src_prop, dst_model_dict in src_prop_dict.items():
            for dst_model, dst_prop_list in dst_model_dict.items():
                for dst_prop_dict in dst_prop_list:
                    dst_prop = next(iter(dst_prop_dict))
                    dst_parents = dst_prop_dict.get(dst_prop).get("Parents", "CONST")
                    cypher_stmts.append(
                        generate_mapping_cypher(
                            src_ent=N(
                                label="property",
                                props={"handle": src_prop, "model": src_model},
                            ),
                            dst_ent=N(
                                label="property",
                                props={"handle": dst_prop, "model": dst_model},
                            ),
                            src_parent=N(
                                props={
                                    "handle": get_parents_as_list(src_parent)[-1],
                                    "model": src_model,
                                }
                            ),
                            dst_parent=N(
                                props={
                                    "handle": get_parents_as_list(dst_parents)[-1],
                                    "model": dst_model,
                                }
                            ),
                            parent_child_rel=R(Type="has_property"),
                            mapping_source=src_model,
                            _commit=_commit,
                        )
                    )
    return cypher_stmts


def convert_mappings_to_changelog(
    mapping_mdf: str, config_file_path: str, author: str, _commit: Optional[str] = None
) -> Changelog:
    """converts mapping MDF to liquibase changelog"""
    changeset_id = changeset_id_generator(config_file_path=config_file_path)
    changelog = Changelog()

    mapping_dict = load_yaml(mapping_mdf)
    cypher_stmts = process_props(mapping_dict, _commit)
    for stmt in cypher_stmts:
        changelog.add_changeset(
            Changeset(
                id=str(next(changeset_id)),
                author=author,
                change_type=CypherChange(text=str(stmt)),
            )
        )

    update_config_changeset_id(
        config_file_path=config_file_path, new_changeset_id=next(changeset_id)
    )

    return changelog


@click.command()
@click.option(
    "--mapping_mdf",
    required=True,
    type=click.Path(exists=True),
    prompt=True,
    help="Mapping MDF File to be converted to changelog",
)
@click.option(
    "--output_file_path",
    required=True,
    type=click.Path(),
    prompt=True,
    help="File path for output changelog",
)
@click.option(
    "--config_file_path",
    required=True,
    type=click.Path(exists=True),
    help="Changeset config file path",
)
@click.option(
    "--author",
    required=True,
    type=str,
    help="author for changeset",
)
@click.option(
    "--_commit",
    required=False,
    type=str,
    help="commit string",
)
def main(
    mapping_mdf: str,
    output_file_path: str,
    config_file_path: str,
    author: str,
    _commit: str,
) -> None:
    """main function to convert mapping mdf file to changelog"""
    changelog = convert_mappings_to_changelog(
        mapping_mdf=mapping_mdf,
        config_file_path=config_file_path,
        author=author,
        _commit=_commit,
    )
    changelog.save_to_file(str(Path(output_file_path)), encoding="UTF-8")


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""
Command line interface for python script to generates tsv file comparing
terms in value sets of properties between 2 models in the MDB.
"""

from pathlib import Path

import click
import pandas as pd
from bento_meta.mdb.mdb_tools.mdb_tools import ToolsMDB
from bento_meta.objects import Property


@click.command()
@click.option(
    "--mdb_uri", required=True, type=str, prompt=True, help="metamodel database URI"
)
@click.option(
    "--mdb_user",
    required=True,
    type=str,
    prompt=True,
    help="metamodel database username",
)
@click.option(
    "--mdb_pass",
    required=True,
    type=str,
    prompt=True,
    help="metamodel database password",
)
@click.option(
    "--base_model",
    required=True,
    default=False,
    type=str,
    prompt=True,
    help="model for value set comparison (e.g., 'PDC')",
)
@click.option(
    "--comp_model",
    required=True,
    default=False,
    type=str,
    prompt=True,
    help="comparison model (e.g., 'GDC')",
)
@click.option(
    "--output_filepath",
    required=True,
    default=False,
    type=str,
    prompt=True,
    help="path to .csv file for output",
)
def main(mdb_uri, mdb_user, mdb_pass, base_model, comp_model, output_filepath):
    """
    Generates tsv file comparing terms in value sets of properties
    between 2 models in the MDB.

    :param str mdb_uri: metamodel database URI
    :param str mdb_user: metamodel database username
    :param str mdb_pass: metamodel database password
    :param str base_model: model for value set comparison (e.g., "PDC")
    :param str comp_model: comparison model (e.g., "GDC")
    :param str output_filepath: path to .tsv file for output
    """
    filepath = Path(output_filepath)
    mdb = ToolsMDB(uri=mdb_uri, user=mdb_user, password=mdb_pass)

    # gather all properties w/ val_sets and their terms for models of interest
    # these are lists of dicts where the keys to each dict are 'prop' & 'terms'
    base_props_terms = mdb.get_props_and_terms_by_model(base_model)
    comp_props_terms = mdb.get_props_and_terms_by_model(comp_model)

    # get list of all nanoids in comp_props_terms
    comp_nanos = []
    for prop_term in comp_props_terms:
        comp_nanos.append(prop_term["prop"]["nanoid"])

    # instantiate list of rows for output
    row_list = []

    # iterate through base props, find synonyms, add terms and number shared between them
    for pt_dict in base_props_terms:
        # set base row to only base prop handle, template for new syn rows
        base_row = {
            "base_prop": pt_dict["prop"]["handle"],
            "synonym_prop": "",
            "shared_terms": "",
            "num_shared_terms": "",
            "base_terms_only": "",
            "synonym_terms_only": "",
        }
        prop = Property({"nanoid": pt_dict["prop"]["nanoid"]})

        # get set of all base prop term values
        base_term_val_set = {str(x["value"]) for x in pt_dict["terms"]}

        # get synonyms for each prop
        prop_synonyms = mdb.get_property_synonyms_all(prop)

        if not prop_synonyms:
            # add row with just base_handle and base terms and move on
            row = base_row
            # add base data to row and append to list of rows (no synonyms)
            row["base_terms_only"] = "|".join(sorted(base_term_val_set))
            row[
                "num_shared_terms"
            ] = f"{str(0)}/{str(len(base_term_val_set))} (both/{base_model})"
            row_list.append(row)
            continue

        # check if synonyms in comp_props_terms
        syn_nanos = []
        for synonym in prop_synonyms:
            # add syn_nano if it is in comp model's nano list and not duplicated
            if synonym.nanoid in comp_nanos and synonym.nanoid not in syn_nanos:
                syn_nanos.append(synonym.nanoid)

        # add row to dataframe with synonym info for each synonym
        for nano in syn_nanos:
            # set row to base_row and add information for base and synonym props/terms
            row = base_row

            # get each synonym's prop and term information
            comp_pt_dict = comp_props_terms[comp_nanos.index(nano)]

            # get set of all comp prop term values
            comp_term_val_set = {str(x["value"]) for x in comp_pt_dict["terms"]}

            # compare sets of terms for base and comp
            terms_both = list(base_term_val_set & comp_term_val_set)
            terms_base_only = list(base_term_val_set - comp_term_val_set)
            terms_comp_only = list(comp_term_val_set - base_term_val_set)

            # add data to row and append to list of rows
            row["synonym_prop"] = comp_pt_dict["prop"]["handle"]
            row["shared_terms"] = "|".join(sorted(terms_both))
            row[
                "num_shared_terms"
            ] = f"{str(len(terms_both))}/{str(len(base_term_val_set))} (both/{base_model})"
            row["base_terms_only"] = "|".join(sorted(terms_base_only))
            row["synonym_terms_only"] = "|".join(sorted(terms_comp_only))
            row_list.append(row)

    # create dataframe from list of row dicts
    output_df = pd.DataFrame.from_records(row_list)

    output_df = output_df.rename(
        columns={
            "base_prop": f"{base_model}_property",
            "synonym_prop": f"{comp_model}_property",
            "shared_terms": "shared_terms",
            "num_shared_terms": "num_shared_terms",
            "base_terms_only": f"{base_model}_terms_only",
            "synonym_terms_only": f"{comp_model}_terms_only",
        }
    )

    # drop any duplicate rows
    output_df = output_df.drop_duplicates()

    # output tsv with results
    output_df.to_csv(filepath, index=False, sep="\t")


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""
Script takes one MDF file representing a model and produces a Liquibase 
Changelog with the necessary changes to add the model to an MDB
"""
import configparser
import logging
from pathlib import Path
from typing import Dict, Generator, List, Optional, Union

import click
from bento_mdf.mdf import MDF
from bento_meta.entity import Entity
from bento_meta.mdb.mdb import make_nanoid
from bento_meta.model import Model
from bento_meta.objects import Concept, Term, ValueSet
from bento_meta.util.cypher.clauses import Create, Match, Merge, OnCreateSet, Statement
from bento_meta.util.cypher.entities import N, R, T, _plain_var
from liquichange.changelog import Changelog, Changeset, CypherChange

logger = logging.getLogger(__name__)


def get_initial_changeset_id(config_file_path: str) -> int:
    """Gets initial changeset id from changelog config file"""
    config = configparser.ConfigParser()
    config.read(config_file_path)
    try:
        return config.getint(section="changelog", option="changeset_id")
    except (configparser.NoSectionError, configparser.NoOptionError) as error:
        logger.error(f"Reading changeset ID failed: {error}")
        raise


def changeset_id_generator(config_file_path: str) -> Generator[int, None, None]:
    """Generates sequential changeset IDs by reading the latest ID from a config file."""
    i = get_initial_changeset_id(config_file_path)
    while True:
        yield i
        i += 1


def update_config_changeset_id(config_file_path: str, new_changeset_id: int) -> None:
    """Updates changelog config file with new changeset id."""
    config = configparser.ConfigParser()
    config.read(config_file_path)
    config.set(section="changelog", option="changeset_id", value=str(new_changeset_id))
    with open(file=config_file_path, mode="w", encoding="UTF-8") as config_file:
        config.write(fp=config_file)


def escape_quotes_in_attr(entity: Entity) -> None:
    """
    Escapes quotes in entity attributes.

    Quotes in string attributes may or may not already be escaped, so this function
    unescapes all previously escaped ' and " characters and replaces them with
    """
    for key, val in vars(entity).items():
        if (
            val
            and val is not None
            and key != "pvt"
            and isinstance(
                val,
                str,
            )
        ):
            # First unescape any previously escaped quotes
            val = val.replace(r"\'", "'").replace(r"\"", '"')

            # Escape all quotes
            val = val.replace("'", r"\'").replace('"', r"\"")

            # Update the modified value back to the attribute
            setattr(entity, key, val)


def cypherize_entity(entity: Entity) -> N:
    """Represents metamodel Entity object as a property graph Node."""
    return N(label=entity.get_label(), props=entity.get_attr_dict())


def generate_cypher_to_add_entity(
    entity: Entity,
    cypher_stmts: Dict[str, List],
) -> None:
    """Generates cypher statement to create or merge Entity."""
    if entity in cypher_stmts["added_entities"]:
        return
    escape_quotes_in_attr(entity)
    cypher_ent = cypherize_entity(entity)
    if isinstance(entity, (Term, ValueSet)):
        if "_commit" not in cypher_ent.props:
            cypher_stmts["add_ents"].append(Statement(Merge(cypher_ent)))
        # remove _commit prop of Term/VS cypher_ent for Merge
        else:
            commit = cypher_ent.props.pop("_commit")
            cypher_stmts["add_ents"].append(
                Statement(Merge(cypher_ent), OnCreateSet(commit))
            )
    else:
        cypher_stmts["add_ents"].append(Statement(Create(cypher_ent)))
    cypher_stmts["added_entities"].append(entity)


def generate_cypher_to_add_relationship(
    src: Entity,
    rel: str,
    dst: Entity,
    cypher_stmts: Dict[str, List],
) -> None:
    """Generates cypher statement to create relationship from src to dst entity"""
    cypher_src = cypherize_entity(src)
    cypher_dst = cypherize_entity(dst)
    # remove _commit attr from Term and VS ents
    for cypher_ent in (cypher_src, cypher_dst):
        if isinstance(cypher_ent, (Term, ValueSet)) and "_commit" in cypher_ent.props:
            cypher_ent.props.pop("_commit")
    cypher_stmts["add_rels"].append(
        Statement(
            Match(cypher_src, cypher_dst),
            Merge(T(_plain_var(cypher_src), R(Type=rel), _plain_var(cypher_dst))),
        )
    )


def process_tags(entity: Entity, cypher_stmts) -> None:
    """Generates cypher statements to create/merge an entity's tag attributes."""
    if not entity.tags:
        return
    for tag in entity.tags.values():
        tag.nanoid = make_nanoid()
        generate_cypher_to_add_entity(tag, cypher_stmts)
        generate_cypher_to_add_relationship(entity, "has_tag", tag, cypher_stmts)


def process_origin(entity: Entity, cypher_stmts) -> None:
    """Generates cypher statements to create/merge an entity's origin attribute."""
    if not entity.origin:
        return
    generate_cypher_to_add_entity(entity.origin, cypher_stmts)
    generate_cypher_to_add_relationship(
        entity, "has_origin", entity.origin, cypher_stmts
    )
    process_tags(entity.origin, cypher_stmts)


def process_terms(entity: Entity, cypher_stmts) -> None:
    """Generates cypher statements to create/merge an entity's term attributes."""
    if not entity.terms:
        return
    for term in entity.terms.values():
        generate_cypher_to_add_entity(term, cypher_stmts)
        if isinstance(entity, Concept):
            generate_cypher_to_add_relationship(
                term, "represents", entity, cypher_stmts
            )
        else:
            generate_cypher_to_add_relationship(entity, "has_term", term, cypher_stmts)
        process_tags(term, cypher_stmts)
        process_origin(term, cypher_stmts)


def process_concept(entity: Entity, cypher_stmts) -> None:
    """Generates cypher statements to create/merge an entity's concept attribute."""
    if not entity.concept:
        return
    generate_cypher_to_add_entity(entity.concept, cypher_stmts)
    generate_cypher_to_add_relationship(
        entity, "has_concept", entity.concept, cypher_stmts
    )
    process_tags(entity.concept, cypher_stmts)
    process_terms(entity.concept, cypher_stmts)


def process_value_set(entity: Entity, cypher_stmts) -> None:
    """Generates cypher statements to create/merge an entity's value_set attribute."""
    if not entity.value_set:
        return
    generate_cypher_to_add_entity(entity.value_set, cypher_stmts)
    generate_cypher_to_add_relationship(
        entity, "has_value_set", entity.value_set, cypher_stmts
    )
    process_tags(entity.value_set, cypher_stmts)
    process_origin(entity.value_set, cypher_stmts)
    process_terms(entity.value_set, cypher_stmts)


def process_props(entity: Entity, cypher_stmts) -> None:
    """Generates cypher statements to create/merge an entity's props attribute."""
    if not entity.props:
        return
    for prop in entity.props.values():
        generate_cypher_to_add_entity(prop, cypher_stmts)
        generate_cypher_to_add_relationship(entity, "has_property", prop, cypher_stmts)
        process_tags(prop, cypher_stmts)
        process_concept(prop, cypher_stmts)
        process_value_set(prop, cypher_stmts)


def process_model_nodes(model: Model, cypher_stmts) -> None:
    """Generates cypher statements to create/merge an model's nodes."""
    for node in model.nodes.values():
        generate_cypher_to_add_entity(node, cypher_stmts)
        process_tags(node, cypher_stmts)
        process_concept(node, cypher_stmts)
        process_props(node, cypher_stmts)


def process_model_edges(model: Model, cypher_stmts) -> None:
    """Generates cypher statements to create/merge an model's edges."""
    for edge in model.edges.values():
        generate_cypher_to_add_entity(edge, cypher_stmts)
        generate_cypher_to_add_relationship(edge, "has_src", edge.src, cypher_stmts)
        generate_cypher_to_add_relationship(edge, "has_dst", edge.dst, cypher_stmts)
        process_tags(edge, cypher_stmts)
        process_concept(edge, cypher_stmts)
        process_props(edge, cypher_stmts)


def convert_model_to_changelog(
    model: Model, author: str, config_file_path: str
) -> Changelog:
    """
    Converts a bento meta model to a Liquibase Changelog.

    Parameters:
    model (Model): The bento meta model to convert
    author (str): The author for the changesets
    config_file_path (str): The path to the changeset config file

    Returns:
    Changelog: The generated Liquibase Changelog

    Functionality:
    - Generates Cypher statements to add entities and relationships from the model
    - Appends Changesets with the Cypher statements to a Changelog
    - Updates the changeset ID in the config file after generating each Changeset
    - Returns the completed Changelog
    """

    cypher_stmts: Dict[str, List] = {
        "add_ents": [],
        "add_rels": [],
        # cypher_stmts["added_entities"] used to prevent duplicate props/terms
        # from being created when they are shared by >1 node/prop
        "added_entities": [],
    }

    # track created props and terms to prevent duplication

    process_model_nodes(model, cypher_stmts)
    process_model_edges(model, cypher_stmts)

    changeset_id = changeset_id_generator(config_file_path=config_file_path)
    changelog = Changelog()

    for stmt in cypher_stmts["add_ents"] + cypher_stmts["add_rels"]:
        changelog.add_changeset(
            Changeset(
                id=str(next(changeset_id)),
                author=author,
                change_type=CypherChange(text=str(stmt)),
            )
        )

    update_config_changeset_id(
        config_file_path=config_file_path, new_changeset_id=next(changeset_id)
    )

    return changelog


@click.command()
@click.option(
    "--model_handle",
    required=True,
    type=str,
    prompt=True,
    help="CRDC Model Handle (e.g. 'GDC')",
)
@click.option(
    "--mdf_files",
    required=True,
    type=click.Path(exists=True),
    prompt=True,
    multiple=True,
    help="MDF file(s)",
)
@click.option(
    "--output_file_path",
    required=True,
    type=click.Path(),
    prompt=True,
    help="File path for output changelog",
)
@click.option(
    "--config_file_path",
    required=True,
    type=click.Path(exists=True),
    help="Changeset config file path",
)
@click.option(
    "--author",
    required=True,
    type=str,
    help="author for changeset",
)
@click.option(
    "--_commit",
    required=False,
    type=str,
    help="commit string",
)
def main(
    model_handle: str,
    mdf_files: Union[str, List[str]],
    output_file_path: str,
    config_file_path: str,
    author: str,
    _commit: Optional[str],
) -> None:
    """get liquibase changelog from mdf files for a model"""
    logger.info("Script started")

    mdf = MDF(*mdf_files, handle=model_handle, _commit=_commit, raiseError=True)
    if not mdf.model:
        raise RuntimeError("Error getting model from MDF")
    logger.info("Model MDF loaded successfully")

    changelog = convert_model_to_changelog(
        model=mdf.model, author=author, config_file_path=config_file_path
    )
    logger.info("Changelog converted from MDF successfully")

    changelog.save_to_file(str(Path(output_file_path)), encoding="UTF-8")
    logger.info("Changelog saved at: {output_file_path}")


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""
bento_meta.util.cypher.clauses

Representations of Cypher statement clauses, statements,
and statement parameters.
"""
from pdb import set_trace
from string import Template

from bento_meta.util.cypher.entities import N, P, R, _condition, _pattern, _return
from bento_meta.util.cypher.functions import Func


class Clause(object):
    """Represents a generic Cypher clause."""

    template = Template("$slot1")
    joiner = ", "

    @staticmethod
    def context(arg):
        return _return(arg)

    def __init__(self, *args, **kwargs):
        self.args = list(args)
        self.kwargs = kwargs

    def __str__(self):
        values = []
        for c in [self.context(x) for x in self.args]:
            if isinstance(c, str):
                values.append(c)
            elif isinstance(c, Func):
                values.append(str(c))
            elif isinstance(c, list):
                values.extend([str(x) for x in c])
            else:
                values.append(str(c))
        return self.template.substitute(slot1=self.joiner.join(values))


class Match(Clause):
    """Create a MATCH clause with the arguments."""

    template = Template("MATCH $slot1")

    @staticmethod
    def context(arg):
        return _pattern(arg)

    def __init__(self, *args):
        super().__init__(*args)


class Where(Clause):
    """Create a WHERE clause with the arguments
    (joining conditions with 'op')."""

    template = Template("WHERE $slot1")
    joiner = " {} "

    @staticmethod
    def context(arg):
        return _condition(arg)

    def __init__(self, *args, op="AND"):
        super().__init__(*args, op=op)
        self.op = op

    def __str__(self):
        values = []
        for c in [self.context(x) for x in self.args]:
            if isinstance(c, str):
                values.append(c)
            elif isinstance(c, Func):
                values.append(str(c))
            elif isinstance(c, list):
                values.extend([str(x) for x in c])
            else:
                values.append(str(c))
        return self.template.substitute(slot1=self.joiner.format(self.op).join(values))


class With(Clause):
    """Create a WITH clause with the arguments."""

    template = Template("WITH $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class Create(Clause):
    """Create a CREATE clause with the arguments."""

    template = Template("CREATE $slot1")

    @staticmethod
    def context(arg):
        return _pattern(arg)

    def __init__(self, *args):
        super().__init__(*args)


class Merge(Clause):
    """Create a MERGE clause with the arguments."""

    template = Template("MERGE $slot1")

    @staticmethod
    def context(arg):
        return _pattern(arg)

    def __init__(self, *args):
        super().__init__(*args)


class Remove(Clause):
    """Create a REMOVE clause with the arguments."""

    template = Template("REMOVE $slot1")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        ent = self.args[0]
        item = ""
        sep = ""
        if "prop" in self.kwargs:
            item = self.kwargs["prop"]
            sep = "."
        elif "label" in self.kwargs:
            item = self.kwargs["label"]
            sep = ":"
        return self.template.substitute(
            slot1="{}{}{}".format(self.context(ent), sep, item)
        )


class Set(Clause):
    """
    Create a SET clause with the arguments. (Only property arguments matter.)
    """

    template = Template("SET $slot1")

    @staticmethod
    def context(arg):
        return _condition(arg)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        values = []
        for c in [self.context(x) for x in self.args if isinstance(x, P)]:
            if isinstance(c, str):
                values.append(c)
            elif isinstance(c, Func):
                values.append(str(c))
            elif isinstance(c, list):
                values.extend([str(x) for x in c])
            else:
                values.append(str(c))
        if "update" in self.kwargs:
            values = [x.replace("=", "+=") for x in values]
        return self.template.substitute(slot1=self.joiner.join(values))


class OnCreateSet(Set):
    """Create an ON CREATE SET clause for a MERGE with the arguments."""

    template = Template("ON CREATE SET $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class OnMatchSet(Set):
    """Create an ON CREATE SET clause for a MERGE with the arguments."""

    template = Template("ON MATCH SET $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class Return(Clause):
    """Create a RETURN clause with the arguments."""

    template = Template("RETURN $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class OptionalMatch(Clause):
    """Create an OPTIONAL MATCH clause with the arguments."""

    template = Template("OPTIONAL MATCH $slot1")

    @staticmethod
    def context(arg):
        return _pattern(arg)

    def __init__(self, *args):
        super().__init__(*args)


class Collect(Clause):
    """Create a COLLECT clause with the arguments."""

    template = Template("COLLECT $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class Unwind(Clause):
    """Create an UNWIND clause with the arguments."""

    template = Template("UNWIND $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class As(Clause):
    """Create an AS clause with the arguments."""

    template = Template("AS $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class Delete(Clause):
    """Create a DELETE clause with the arguments."""

    template = Template("DELETE $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class DetachDelete(Clause):
    """Create a DETACH DELETE clause with the arguments."""

    template = Template("DETACH DELETE $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class Case(Clause):
    """Create a CASE clause with the arguments."""

    template = Template("CASE $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class When(Clause):
    """Create a WHEN clause with the arguments."""

    template = Template("WHEN $slot1")
    joiner = " {} "

    @staticmethod
    def context(arg):
        return _condition(arg)

    def __init__(self, *args, op="AND"):
        super().__init__(*args, op=op)
        self.op = op

    def __str__(self):
        values = []
        for c in [self.context(x) for x in self.args]:
            if isinstance(c, str):
                values.append(c)
            elif isinstance(c, Func):
                values.append(str(c))
            elif isinstance(c, list):
                values.extend([str(x) for x in c])
            else:
                values.append(str(c))
        return self.template.substitute(slot1=self.joiner.format(self.op).join(values))


class Then(Clause):
    """Create a THEN clause with the arguments."""

    template = Template("THEN $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class Else(Clause):
    """Create an ELSE clause with the arguments."""

    template = Template("ELSE $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class End(Clause):
    """Create an END clause with the arguments."""

    template = Template("END $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class ForEach(Clause):
    """Create an FOREACH clause with the arguments."""

    template = Template("FOREACH $slot1")

    def __init__(self, *args):
        super().__init__(*args)


class Statement(object):
    """Create a Neo4j statement comprised of clauses (and strings) in order."""

    def __init__(self, *args, terminate=False, use_params=False):
        self.clauses = args
        self.terminate = terminate
        self.use_params = use_params
        self._params = None

    def __str__(self):
        stash = P.parameterize
        if self.use_params:
            P.parameterize = True
        else:
            P.parameterize = False
        ret = " ".join([str(x) for x in self.clauses])
        if self.terminate:
            ret = ret + ";"
        P.parameterize = stash
        return ret

    # def params(self):
    #     if not self._params:
    #         self._params = re.findall("\\$[a-zA-A0-9_]+", str(self))
    #     return self._params
    @property
    def params(self):
        if self._params is None:
            self._params = {}
            for c in self.clauses:
                for ent in c.args:
                    if isinstance(ent, (N, R)):
                        for p in ent.props.values():
                            self._params[p.var] = p.value
                    else:
                        if "nodes" in vars(type(ent)):
                            for n in ent.nodes():
                                for p in n.props.values():
                                    self._params[p.var] = p.value
                        if "edges" in vars(type(ent)):
                            for e in ent.edges():
                                for p in e.props.values():
                                    self._params[p.var] = p.value
        return self._params



"""
mdb.searchable
Subclass of `class:bento_meta.mdb.MDB` to support searching fulltext indices on an MDB
Note: certain fulltext indexes on certain MDB nodes and properties must be present in 
the Neo4j instance: 
- entityHandles
- termValue
- termDefn
- termValueDefn

"""
from bento_meta.mdb import read_txn, read_txn_data, read_txn_value
from bento_meta.mdb import MDB
from pdb import set_trace
class SearchableMDB(MDB):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ftindexes = {}
        with self.driver.session() as s:
            result = s.run("call db.indexes")
            for rec in result:
                if rec['type'] == 'FULLTEXT':
                    self.ftindexes[rec['name']] = {
                        "entity_type":rec['entityType'],
                        "entities":rec['labelsOrTypes'],
                        "properties":rec['properties'],
                    }

    def available_indexes(self):
        """Fulltext indexes present in database. 
        Returns { <index_name> : { entity_type:<NODE|RELATIONSHIP>, entities:[<labels>], 
        properties:[ [<props>] ] } } """
        return self.ftindexes

    @read_txn_data
    def query_index(self, index, qstring, skip=None, limit=None):
        """Query a named fulltext index of nodes or relationships.
        Returns [ {ent:{}, label:<label>, score:<lucene score>} ]."""
        if index not in self.ftindexes:
            raise RuntimeError("Index with name '{}' not found".format(index))
        tipe = ""
        if self.ftindexes[index]['entity_type'] == "NODE":
            tipe = "queryNodes"
        elif self.ftindexes[index]['entity_type'] == "RELATIONSHIP":
            tipe = "queryRelationships"
        else:
            raise RuntimeError("Wha?")
        thing = self.ftindexes[index]['entity_type'].lower()
        qry = (
            "call db.index.fulltext.{tipe}($name, $query) "
            "yield {thing}, score "
            "return {thing} as ent, head(labels({thing})) as label, score "
            "{skip} {limit} "
            ).format(tipe=tipe, thing=thing, skip="SKIP $skip" if skip else "",
                     limit="LIMIT $limit" if limit else "")
        parms = {"name": index, "query": qstring}
        if skip:
            parms['skip'] = skip
        if limit:
            parms['limit'] = limit
        return (qry, parms)

    def search_entity_handles(self, qstring):
        """Fulltext search of qstring over node, relationship, and property handles.
        Returns { node:[ {ent:<entity dict>,score:<lucene score>},... ],
        relationship:[ <...> ], property:[ <...> ] }"""

        result = self.query_index('entityHandle', qstring)
        if not result:
            return None
        plural = {"node":"nodes", "relationship":"relationships",
                  "property":"properties"}
        ret = {"nodes": [], "relationships": [], "properties": []}
        for item in result:
            ret[plural[item['label']]].append({"ent": item['ent'], "score": item['score']}) 
        return ret

    def search_terms(self, qstring, search_values=True,
                     search_definitions=True):
        """Fulltext for qstring over terms, by value, definition, or both (default).
        Returns [ { ent:<term dict>, score:<lucene score> } ]}"""
        index = {True: {True: 'termValueDefn', False: 'termDefn'},
                 False: {True: 'termValue', False: None}}
        result = self.query_index(index[search_definitions][search_values],
                                  qstring)
        return result



"""
bento_meta.mdb
==============

This module contains :class:`MDB`, with machinery for efficiently
querying a Neo4j instance of a Metamodel Database.
"""

from .mdb import (
    MDB, make_nanoid, read_txn,
    read_txn_value, read_txn_data
    )
from .writeable import WriteableMDB
from .searchable import SearchableMDB
from .loaders import load_mdf, load_model, load_model_statements




# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bento_meta',
 'bento_meta.mdb',
 'bento_meta.mdb.mdb_tools',
 'bento_meta.util',
 'bento_meta.util.cypher']

package_data = \
{'': ['*'], 'bento_meta': ['logs/*']}

install_requires = \
['PyYAML>=6.0.1',
 'bento-mdf>=0.9.1,<0.10.0',
 'delfick-project>=0.7.9,<0.8.0',
 'nanoid>=2.0.0,<3.0.0',
 'neo4j>=4.0',
 'requests>=2.28.1,<3.0.0',
 'setuptools>=65.4.1,<66.0.0',
 'tqdm>=4.64.1,<5.0.0']

extras_require = \
{':extra == "tools"': ['liquichange>=0.2.1,<0.3.0'],
 'tools': ['numpy>=1.23.5,<2.0.0',
           'pandas>=1.5.2,<2.0.0',
           'spacy>=3.4.3,<4.0.0',
           'click>=8.1.3,<9.0.0']}

scripts = \
['scripts/compare_models.py',
 'scripts/compare_val_set_terms.py',
 'scripts/make_model_changelog.py',
 'scripts/make_diff_changelog.py',
 'scripts/make_mapping_changelog.py']

setup_kwargs = {
    'name': 'bento-meta',
    'version': '0.2.3',
    'description': 'Python drivers for Bento Metamodel Database',
    'long_description': '# bento_meta - Object Model for the Bento Metamodel Database\n\nRead the docs at\n[https://cbiit.github.io/bento-meta/](https://cbiit.github.io/bento-meta/).\n\n\n',
    'author': 'Mark A. Jensen',
    'author_email': 'mark.jensen@nih.gov',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'scripts': scripts,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
bento_meta.mdb
==============

This module contains :class:`MDB`, with machinery for efficiently
querying a Neo4j instance of a Metamodel Database.
"""
import os
import re
from warnings import warn
from functools import wraps
from neo4j import GraphDatabase
from nanoid import generate as nanoid_generate
from pdb import set_trace

# Decorator functions to produce executed transactions based on an
# underlying query/param function:

def read_txn(func):
    """Decorates a query function to run a read transaction based on
    its query.
    Query function should return a tuple (qry_string, param_dict).
    Returns list of driver Records."""
    @wraps(func)
    def rd(self, *args, **kwargs):
        def txn_q(tx):
            (qry,parms)=func(self, *args, **kwargs)
            result = tx.run(qry,parameters=parms)
            return [rec for rec in result]
        with self.driver.session() as session:
            result = session.read_transaction(txn_q)
            return result
    return rd

def read_txn_value(func):
    """Decorates a query function to run a read transaction based on
    its query.
    Query function should return a tuple (qry_string, param_dict, values_key).
    Returns list of values for key specified by query function."""
    @wraps(func)
    def rd(self, *args, **kwargs):
        def txn_q(tx):
            (qry,parms,values_key)=func(self, *args, **kwargs)
            result = tx.run(qry,parameters=parms)
            return result.value(values_key)
        with self.driver.session() as session:
            result = session.read_transaction(txn_q)
            return result
    return rd

def read_txn_data(func):
    """Decorates a query function to run a read transaction based on
    its query.
    Query function should return a tuple (qry_string, param_dict).
    Returns records as a list of simple dicts."""
    @wraps(func)
    def rd(self, *args, **kwargs):
        (qry,parms)=func(self, *args, **kwargs)
        def txn_q(tx):
            result = tx.run(qry,parameters=parms)
            return result.data()
        with self.driver.session() as session:
            result = session.read_transaction(txn_q)
            if len(result):
                return result
            else:
                return None
    return rd

class MDB:
    def __init__(self, uri=os.environ.get("NEO4J_MDB_URI"),
                 user=os.environ.get("NEO4J_MDB_USER"),
                 password=os.environ.get("NEO4J_MDB_PASS")):
        self.uri = uri
        self.user = user
        self.password = password
        self.driver = None
        try:
            self.driver = GraphDatabase.driver(self.uri,
                                               auth=(self.user, self.password))
        except Exception as e:
            warn("MDB not connected: {}".format(e))
        self._txfns = {}
        """ Create an :class:`MDB` object, with a connection to a Neo4j instance of a metamodel database.
        :param bolt_url uri: The Bolt protocol endpoint to the Neo4j instance (default, use the
        ``NEO4J_MDB_URI`` env variable)
        :param str user: Username for Neo4j access (default, use the ``NEO4J_MDB_USER`` env variable)
        :param str password: Password for user (default, use the ``NEO4J_MDB_PASS`` env variable)"""
    def close(self):
        self.driver.close()


    def register_txfn(self, name, fn):
        """Register a transaction function
        (see https://neo4j.com/docs/api/python-driver/current/api.html#managed-transactions-transaction-functions)
        with the class for later use."""

        self._txfns[name] = fn

    # def run_txfn(self, name, *args, **kwargs):

    @read_txn_value
    def get_model_handles(self):
        """Return a simple list of model handles available."""
        qry = (
            "match (p:node) "
            "where not exists(p._to) "
            "return distinct p.model"
            )
        return (qry, None,"p.model")

    @read_txn_data
    def get_model_nodes(self, model=None):
        """Return a list of dicts representing Model nodes."""
        qry = (
            "match (m:model) {} "
            "with m "
            "where not exists(m._to) "
            "return m"
            ).format("where m.handle = $model" if model else "")
        return(qry, {"model":model} if model else None)
    
    @read_txn_value
    def get_nodes_by_model(self, model=None):
        """Get all nodes for a given model. If :param:model is None, 
        get all nodes in database.
        Returns [ <node> ]."""
        qry = (
            "match (n:node) {} "
            "with n "
            "where not exists(n._to) "
            "return n"
            ).format("where n.model = $model" if model else "")
        return (qry, {"model": model} if model else None, "n")
    
    @read_txn_data
    def get_model_nodes_edges(self, model):
        """Get all node-relationship-node paths for a given model.
        Returns [ path ]"""
        qry = (
            "match p = (s:node {model: $model})<-[:has_src]-"
            "          (r:relationship {model: $model})-[:has_dst]->"
            "          (d:node {model: $model}) "
            "where not exists(s._to) and not exists(r._to) and "
            "not exists(d._to) "
            "return p as path"
            )
        return (qry, {"model":model})

    @read_txn_data
    def get_node_edges_by_node_id(self, nanoid):
        """Get incoming and outgoing relationship information for a node, 
        given its nanoid.
        Returns [ {id, handle, model, near_type, far_type, rln, far_node} ]."""
        qry = (
            "match (n:node {nanoid:$nanoid}) "
            "where not exists(n._to) "
            "with n "
            "optional match (n)<-[e1]-(r:relationship)-[e2]->(m:node) "
            "where not exists(r._to) and not exists(m._to) "
            "return n.nanoid as id, n.handle as handle, n.model as model, "
            "       type(e1) as near_type, type(e2) as far_type, r as rln, m as far_node"
            )
        return (qry, {"nanoid": nanoid})

    @read_txn_data
    def get_node_and_props_by_node_id(self, nanoid):
        """Get a node and its properties, given the node nanoid.
        Returns [ {id, handle, model, node, props[]} ]."""
        qry = (
            "match (n:node {nanoid:$nanoid}) "
            "where not exists(n._to) "
            "with n "
            "optional match (n)-[:has_property]->(p) "
            "where not exists(p._to) "
            "return n.nanoid as id, n.handle as handle, n.model as model, n as node, "
            "       collect(p) as props"
            )
        return (qry, {"nanoid": nanoid})

    @read_txn_data
    def get_nodes_and_props_by_model(self, model=None):
        """Get all nodes with associated properties given a model handle. If
        model is None, get all nodes with their properties.
        Returns [ {id, handle, model, props[]} ]"""
        qry = (
            "match (n:node)-[:has_property]->(p:property) "
            "where not exists(n._to) and not exists(p._to) {} "
            "return n.nanoid as id, n.handle as handle, n.model as model, "
            "       collect(p) as props"
            ).format("and n.model = $model" if model else "")
        return (qry, {"model":model})
    
            
    @read_txn_data
    def get_prop_node_and_domain_by_prop_id(self, nanoid):
        """Get a property, its node, and its value domain or value set
        of terms, given the property nanoid.
        Returns [ { id, handle, model, value_domain, prop, node, value_set, terms[] } ]."""
        qry = (
            "match (p:property {nanoid:$nanoid})<-[:has_property]-(n:node) "
            "where not exists(p._to) "
            "with p,n "
            "optional match (p)-[:has_value_set]->(vs:value_set)-[:has_term]->(t:term) "
            "return p.nanoid as id, p.handle as handle, p.model as model, "
            "p.value_domain as value_domain, p as prop, n as node, "
            "  vs as value_set, collect(t) as terms"
            )
        return (qry, {"nanoid": nanoid})

    @read_txn_data
    def get_valueset_by_id(self, nanoid):
        """Get a valueset with the properties that use it and the terms
        that constitute it.
        Returns [ {id, handle, url, terms[], props[]} ]"""
        qry = (
            "match (vs:value_set {nanoid:$nanoid}) "
            "with vs "
            "match (t)<-[:has_term]-(vs)<-[:has_value_set]-(p:property) "
            "where not exists(t._to) and not exists(p._to) "
            "return vs.nanoid as id, vs.handle as handle, vs.url as url, "
            "       collect(t) as terms, collect(p) as props"
            )
        return (qry, {"nanoid": nanoid})

    @read_txn_data
    def get_valuesets_by_model(self, model=None):
        """Get all valuesets that are used by properties in the given
        model (or all valuesets if model is None). Also return list of properties using
        each valueset.
        Returns [ {value_set, props[]} ]."""
        qry = (
            "match (vs:value_set)<-[:has_value_set]-(p:property) "
            "where not exists(vs._to) and not exists(p._to) {} "
            "return vs as value_set, collect(p) as props"
            ).format("and p.model=$model" if model else "")
        return (qry, {"model": model})

        
    @read_txn_data
    def get_term_by_id(self, nanoid):
        """Get a term having the given nanoid, with its origin.
        Returns {term, origin}."""
        qry = (
            "match (t:term {nanoid:$nanoid}) "
            "where not exists(t._to) "
            "with t, t.origin_name as origin_name "
            "optional match (o:origin {name: origin_name}) "
            "where not exists(o._to) "
            "return t as term, o as origin "
            )
        return (qry, {"nanoid": nanoid})

    @read_txn_data
    def get_props_and_terms_by_model(self, model=None):
        """Get terms from valuesets associated with properties in a given model
        (or all such terms if model is None).
        Returns [ {prop, terms[]} ]"""
        qry = (
            "match (p:property)-[:has_value_set]->(v:value_set)"
            "-[:has_term]->(t:term) "
            "where not( exists(p._to) or exists(v._to) or exists(t._to)) {} "
            "return p as prop, collect(t) as terms"
            ).format("and p.model = $model" if model else "")
        return (qry, {"model": model})

    @read_txn_data
    def get_origins(self):
        """Get all origins.
        Returns [ <origin> ]"""
        qry = (
            "match (o:origin) "
            "where not exists (o._to) "
            "return o"
            )
        return (qry, None)

    @read_txn_value
    def get_origin_by_id(self, oid):
        """Get an origin by nanoid."""
        qry = (
            "match (o:origin {nanoid:$oid}) "
            "where not exists (o._to) "
            "return o "
            )
        return (qry, {"oid": oid}, "o")

    @read_txn_data
    def get_tags_for_entity_by_id(self, nanoid):
        """Get all tags attached to an entity, given the entity's nanoid.
        Returns [ {model(str), tags[]} ]."""
        qry = (
            "match (a {nanoid:$nanoid})-[:has_tag]->(g:tag) "
            "where not exists(a._to) "
            "with a, g "
            "return a.nanoid as id, head(labels(a)) as label, a.handle as handle, "
            "a.model as model, collect(g) as tags"
            )
        return (qry, {"nanoid":nanoid})

    @read_txn_data
    def get_tags_and_values(self, key=None):
        """Get all tag key/value pairs that are present in database.
        Returns [ { key(str) : values[] } ]"""
        qry = (
            "match (t:tag) {} "
            "return t.key as key, collect(distinct t.value) as values "
            ).format("where t.key = $key" if key else "")
        return (qry, {"key": key} if key else {})
        

    @read_txn_data
    def get_entities_by_tag(self, key, value=None, model=None):
        """Get all entities, optionally from a given model,
        tagged with a given key or key:value pair.
        Returns [ {tag_key(str), tag_value(str), entity(str - label), entities[]} ]"""
        qry = (
            "match (t:tag {{key:$key}}) {} "
            "with t "
            "match (e)-[:has_tag]->(t) "
            "where not exists(e._to) {}"
            "return t.key as tag_key, t.value as tag_value, head(labels(e)) as entity, collect(e) as entities"
            ).format("where t.value = $value" if value else "",
                         "and e.model = $model " if model else "" )
        parms = {"key":key}
        if value:
            parms["value"]=value
        if model:
            parms["model"]=model
        return (qry, parms)

    @read_txn_data
    def get_with_statement(self, qry, parms={}):
        """Run an arbitrary read statement and return data."""
        if not isinstance(qry, str):
            raise RuntimeError("qry= must be a string")
        if not re.match(".*return.*", qry, flags=re.I):
            raise RuntimeError("Read statement needs a RETURN clause.")
        if not isinstance(parms, dict):
            raise RuntimeError("parms= must be a dict")
        return (qry, parms)
    
def make_nanoid(alphabet="abcdefghijkmnopqrstuvwxyzABCDEFGHJKMNPQRSTUVWXYZ0123456789",
           size=6):
    """Create a random nanoid and return it as a string."""
    return nanoid_generate(alphabet=alphabet, size=size)
